"""
this module is using option-chain-analytics package
to fetch OptionChain data with options data
see https://pypi.org/project/option-chain-analytics
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from qis import TimePeriod
from typing import Dict, Tuple, Optional, Literal
from numba.typed import List
from enum import Enum
import qis as qis
# chain
from option_chain_analytics import OptionsDataDFs, create_chain_from_from_options_dfs
from option_chain_analytics.option_chain import SliceColumn, SlicesChain
# analytics
from stochvolmodels.data.option_chain import OptionChain


def generate_vol_chain_np(chain: SlicesChain,
                          value_time: pd.Timestamp,
                          days_map: Dict[str, int] = {'1w': 7, '1m': 21},
                          delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                          is_filtered: bool = True
                          ) -> OptionChain:
    """
    given SlicesChain generate OptionChain for calibration inputs
    """

    ttms, future_prices, discfactors = List(), List(), List()
    optiontypes_ttms, strikes_ttms = List(), List()
    bid_ivs, ask_ivs = List(), List()
    bid_prices, ask_prices = List(), List()
    slice_ids = []
    for label, day in days_map.items():
        next_date = value_time + pd.DateOffset(days=day)  # if overlapping next date will be last avilable maturity
        slice_date = chain.get_next_slice_after_date(mat_date=next_date)
        slice_t = chain.expiry_slices[slice_date]
        df = slice_t.get_joint_slice(delta_bounds=delta_bounds, is_filtered=is_filtered)
        if not df.empty:
            slice_ids.append(f"{label}: {slice_t.expiry_id}")
            ttms.append(slice_t.get_ttm())
            future_prices.append(slice_t.get_future_price())
            discfactors.append(1.0)
            strikes_ttms.append(df.index.to_numpy())
            optiontypes_ttms.append(df[SliceColumn.OPTION_TYPE].to_numpy(dtype=str))
            bid_ivs.append(df[SliceColumn.BID_IV].to_numpy())
            ask_ivs.append(df[SliceColumn.ASK_IV].to_numpy())
            bid_prices.append(df[SliceColumn.BID_PRICE].to_numpy())
            ask_prices.append(df[SliceColumn.ASK_PRICE].to_numpy())

    out = OptionChain(ttms=np.array(ttms),
                      forwards=np.array(future_prices),
                      discfactors=np.array(discfactors),
                      ids=np.array(slice_ids),
                      strikes_ttms=strikes_ttms,
                      optiontypes_ttms=optiontypes_ttms,
                      bid_ivs=bid_ivs,
                      ask_ivs=ask_ivs,
                      bid_prices=bid_prices,
                      ask_prices=ask_prices)
    return out


def load_option_chain(options_data_dfs: OptionsDataDFs,
                      value_time: pd.Timestamp = pd.Timestamp('2023-02-06 08:00:00+00:00'),
                      days_map: Dict[str, int] = {'1w': 7, '1m': 21},
                      delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                      is_filtered: bool = True
                      ) -> Optional[OptionChain]:
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
    if chain is not None:
        option_chain = generate_vol_chain_np(chain=chain,
                                             value_time=value_time,
                                             days_map=days_map,
                                             delta_bounds=delta_bounds,
                                             is_filtered=is_filtered)
    else:
        option_chain = None

    return option_chain


def sample_option_chain_at_times(options_data_dfs: OptionsDataDFs,
                                 time_period: TimePeriod,
                                 freq: str = 'W-FRI',
                                 days_map: Dict[str, int] = {'1w': 7, '1m': 21},
                                 delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.1, 0.1),
                                 hour_offset: int = 8
                                 ) -> Dict[pd.Timestamp, OptionChain]:
    value_times = qis.generate_dates_schedule(time_period=time_period,
                                              freq=freq,
                                              hour_offset=hour_offset)
    option_chains = {}
    for value_time in value_times:
        option_chains[value_time] = load_option_chain(options_data_dfs=options_data_dfs,
                                                      value_time=value_time,
                                                      days_map=days_map,
                                                      delta_bounds=delta_bounds,
                                                      is_filtered=True)
    return option_chains


def load_price_data(options_data_dfs: OptionsDataDFs,
                    time_period: TimePeriod = None,
                    data: Literal['close', 'perp', 'funding_rate'] = 'close',
                    freq: Optional[str] = 'D'  # to do
                    ) -> pd.Series:
    #options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, freq='D', hour_offset=8))
    spot_price = options_data_dfs.get_spot_data()[data]
    if freq is not None:
        spot_price = spot_price.resample(freq).last()
    if time_period is not None:
        spot_price = time_period.locate(spot_price)
    return spot_price


class UnitTests(Enum):
    PRINT_CHAIN_DATA = 1
    GENERATE_VOL_CHAIN_NP = 2
    SAMPLE_CHAIN_AT_TIMES = 3


def run_unit_test(unit_test: UnitTests):

    ticker = 'BTC'  # BTC, ETH
    value_time = pd.Timestamp('2021-10-21 08:00:00+00:00')
    value_time = pd.Timestamp('2023-10-06 08:00:00+00:00')

    from option_chain_analytics.ts_loaders import ts_data_loader_wrapper
    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker))
    options_data_dfs.get_start_end_date().print()
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)

    if unit_test == UnitTests.PRINT_CHAIN_DATA:
        for expiry, eslice in chain.expiry_slices.items():
            eslice.print()

    elif unit_test == UnitTests.GENERATE_VOL_CHAIN_NP:
        option_chain = generate_vol_chain_np(chain=chain,
                                             value_time=value_time,
                                             days_map={'1w': 7},
                                             delta_bounds=(-0.1, 0.1),
                                             is_filtered=True)
        option_chain.print()
        skews = option_chain.get_chain_skews(delta=0.35)
        print(skews)

    elif unit_test == UnitTests.SAMPLE_CHAIN_AT_TIMES:
        time_period = qis.TimePeriod('01Jan2023', '31Jan2023', tz='UTC')
        option_chains = sample_option_chain_at_times(options_data_dfs=options_data_dfs,
                                                     time_period=time_period,
                                                     freq='W-FRI',
                                                     hour_offset=9
                                                     )
        for key, chain in option_chains.items():
            print(f"{key}")
            print(chain)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SAMPLE_CHAIN_AT_TIMES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
from dataclasses import dataclass
from typing import Tuple, Optional
from numba.typed import List
from abc import ABC, abstractmethod

from stochvolmodels.pricers.factor_hjm.rate_core import bond, swap_grad


class BasisHJM(ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_basis(self, tau: float) -> np.ndarray:
        """return (functional) main basis for a given tenor tau"""
        pass

    @abstractmethod
    def get_aux_basis(self, tau: float) -> np.ndarray:
        """return (functional) auxiliary basis for a given tenor tau"""
        pass

    @staticmethod
    @abstractmethod
    def get_nb_factors(cls) -> np.ndarray:
        """return number of model factors """
        pass

    @staticmethod
    @abstractmethod
    def get_nb_aux_factors(cls) -> np.ndarray:
        """return number of auxiliary factors """
        pass

    @abstractmethod
    def bond_coeffs(self, tau: float) -> Tuple[np.ndarray, np.ndarray]:
        pass

    @abstractmethod
    def calc_Omega(self, M: np.ndarray) -> np.ndarray:
        pass

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def _bond(self, nb_factors: int, nb_aux_factors: int,
              t: float, T: float,
              x: np.ndarray, y: np.ndarray,
              ccy: str,
              m: int = 0) -> np.ndarray:
        assert t <= T
        # sime of x is # simulations x #factors
        # assume that in matrix X the row is for trials
        # column is for factors
        assert x.shape[-1] == nb_factors and y.shape[-1] == nb_aux_factors
        tau = T - t
        B_PX, B_PY = self.bond_coeffs(tau)
        # there are some other checks at bond function
        bond_value = bond(t, T, x, y, B_PX, B_PY, ccy, m)

        return bond_value

    def _get_matrix_B(self, nb_factors: int, key_terms: np.ndarray) -> np.ndarray:
        nb_key_terms = key_terms.size
        B = np.zeros((nb_key_terms, nb_factors))
        for idx, tau in enumerate(key_terms):
            B[idx, :] = 1.0/tau * self.bond_coeffs(tau)[0]

        return B

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def annuity(self, t: float, ts_sw: np.ndarray,
                x: np.ndarray, y: np.ndarray,
                ccy: str,
                m: int = 0) -> np.ndarray:
        ann = 0
        for i in range(1, ts_sw.size):
            bond_value = self.bond(t, ts_sw[i], x, y, ccy, m)
            ann = ann + (ts_sw[i] - ts_sw[i - 1]) * bond_value
        return ann

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def swap_rate(self, t: float, ts_sw: np.ndarray,
                  x: np.ndarray, y: np.ndarray,
                  ccy: str) -> (np.ndarray, np.ndarray):
        denumer0 = 0
        denumer1 = 0
        for i in range(1, ts_sw.size):
            denumer0 = denumer0 + (ts_sw[i] - ts_sw[i - 1]) * self.bond(t, ts_sw[i], x, y, ccy=ccy, m=0)
            denumer1 = denumer1 + (ts_sw[i] - ts_sw[i - 1]) * self.bond(t, ts_sw[i], x, y, ccy=ccy, m=1)

        numer0 = self.bond(t, ts_sw[0], x, y, ccy=ccy, m=0) - self.bond(t, ts_sw[-1], x, y, ccy=ccy, m=0)
        numer1 = self.bond(t, ts_sw[0], x, y, ccy=ccy, m=1) - self.bond(t, ts_sw[-1], x, y, ccy=ccy, m=1)
        # ad-hoc fix
        # if numer1.ndim == 2:
        #     numer1 = numer1[:, 0]

        value0 = numer0 / denumer0
        value1 = swap_grad(numer0=numer0, numer1=numer1, denumer0=denumer0, denumer1=denumer1)

        return value0, value1

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def libor_rate(self, t: float, t_start: float, t_end: float,
                   x: np.ndarray, y: np.ndarray,
                   ccy: str):
        zcb_start = self.bond(t, t_start, x, y, ccy=ccy, m=0)
        zcb_end = self.bond(t, t_end, x, y, ccy=ccy, m=0)
        libor = 1.0 / (t_end - t_start) * (zcb_start / zcb_end - 1.0)

        return libor

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def calculate_swap_rate(self,
                            ttm: float,
                            x0: np.ndarray,
                            y0: np.ndarray,
                            I0: np.ndarray,
                            ts_sw: np.ndarray,
                            ccy: str) -> List[np.ndarray, ...]:
        # calculate swap rates using simulated states
        # ts_sw = get_default_swap_term_structure(ttm)
        s_mc = self.swap_rate(t=ttm, ts_sw=ts_sw, x=x0, y=y0, ccy=ccy)[0]
        ann_mc = self.annuity(t=ttm, ts_sw=ts_sw, x=x0, y=y0, m=0, ccy=ccy)
        numer = 1.0 / self.bond(t=0, T=ttm, x=np.zeros((1, x0.shape[1])), y=np.zeros((1, y0.shape[1])), m=0, ccy=ccy) * np.exp(I0)
        return s_mc, ann_mc, numer


#############################################################################
#                     Single factor Cheyette model
#############################################################################
@dataclass
class Cheyette1D(BasisHJM):
    meanrev: float

    def __post_init__(self):
        assert self.meanrev > 0
        self.nb_factors = Cheyette1D.get_nb_factors()
        self.nb_aux_factors = Cheyette1D.get_nb_aux_factors()

    def get_basis(self, tau: float) -> np.ndarray:
        raise NotImplementedError(f"not supported for Cheyette1D")

    def get_aux_basis(self, tau: float):
        raise NotImplementedError(f"not supported for Cheyette1D")

    @classmethod
    def get_nb_factors(cls) -> int:
        return 1

    @classmethod
    def get_nb_aux_factors(cls) -> int:
        return 1

    def get_generating_matrix(self) -> np.ndarray:
        # D = np.zeros((self.nb_factors, self.nb_factors))
        # D[1, 1] = D[2, 2] = -self.meanrev
        # D[1, 2] = 1.0
        # return D
        raise NotImplementedError(f"not supported for Cheyette1D")

    def get_aux_generating_matrix(self) -> np.ndarray:
        # D = np.zeros((self.nb_aux_factors, self.nb_aux_factors))
        # D[0, 1] = 1.0
        # D[2, 2] = D[3, 3] = D[4, 4] = -self.meanrev
        # D[2, 3] = D[3, 4] = 1.0
        # D[5, 5] = D[6, 6] = D[7, 7] = -2.0 * self.meanrev
        # D[5, 6] = D[6, 7] = 1.0
        # return D
        raise NotImplementedError(f"not supported for Cheyette1D")

    def get_matrix_B(self, key_terms: List[float]) -> np.ndarray:
        # assert len(key_terms) == 3  # for now we assume three key terms
        # B = np.zeros((len(key_terms), self.nb_factors))
        # for idx, tau in enumerate(key_terms):
        #     B[idx, :] = self.get_basis(tau)
        #
        # return B
        raise NotImplementedError(f"not supported for Cheyette1D")

    def calc_Omega(self, M: np.ndarray) -> np.ndarray:
        raise NotImplementedError(f"not supported for Cheyette1D")

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def bond_coeffs(self, tau: float) -> Tuple[np.ndarray, np.ndarray]:
        G_tau = (1.0 - np.exp(-self.meanrev * tau)) / self.meanrev
        B_PX = np.array([G_tau])
        B_PY = np.array([0.5*G_tau*G_tau])
        return B_PX, B_PY

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def bond(self, t: float, T: float,
             x: np.ndarray, y: np.ndarray,
             ccy: str,
             m: int = 0) -> np.ndarray:
        assert t <= T
        if x.ndim == 1:
            # sime of x is #factors * # simulations
            # assume that in matrix x the columns is for trials
            # rows are for factors
            assert x.shape[0] == y.shape[0]
        tau = T - t
        B_PX, B_PY = self.bond_coeffs(tau)
        # there are some other checks at bond function
        bond_value = bond(t, T, x, y, B_PX, B_PY, ccy, m)
        if x.ndim == 0:
            bond_value = bond_value[0][0]

        return bond_value


#############################################################################
#                     Nelson-Siegel FHJM model
#############################################################################
@dataclass
class NelsonSiegel(BasisHJM):
    meanrev: float
    key_terms: np.ndarray

    def __post_init__(self):
        assert self.meanrev > 0
        self.nb_factors = NelsonSiegel.get_nb_factors()
        self.nb_aux_factors = NelsonSiegel.get_nb_aux_factors()
        # same number of key terms and factors
        assert self.key_terms.size == self.nb_factors


    def get_basis(self, tau: float) -> np.ndarray:
        mrv = self.meanrev
        exp_mrv = np.exp(-mrv * tau)
        return np.array([1.0, exp_mrv, tau * exp_mrv])

    def get_aux_basis(self, tau: float):
        mrv = self.meanrev
        exp_mrv = np.exp(-mrv * tau)
        exp_mrv2 = np.exp(-2.0 * mrv * tau)
        return np.array([1.0, tau,
                         exp_mrv, tau * exp_mrv, 0.5 * tau * tau * exp_mrv,
                         exp_mrv2, tau * exp_mrv2, 0.5 * tau * tau * exp_mrv2])

    @classmethod
    def get_nb_factors(cls) -> int:
        return 3

    @classmethod
    def get_nb_aux_factors(cls) -> int:
        return 8

    def get_generating_matrix(self) -> np.ndarray:
        D = np.zeros((self.nb_factors, self.nb_factors))
        D[1, 1] = D[2, 2] = -self.meanrev
        D[1, 2] = 1.0
        return D

    def get_aux_generating_matrix(self) -> np.ndarray:
        D = np.zeros((self.nb_aux_factors, self.nb_aux_factors))
        D[0, 1] = 1.0
        D[2, 2] = D[3, 3] = D[4, 4] = -self.meanrev
        D[2, 3] = D[3, 4] = 1.0
        D[5, 5] = D[6, 6] = D[7, 7] = -2.0 * self.meanrev
        D[5, 6] = D[6, 7] = 1.0
        return D


    def get_matrix_B(self) -> np.ndarray:
        return self._get_matrix_B(self.nb_factors, self.key_terms)

    def calc_Omega(self, M: np.ndarray) -> np.ndarray:
        assert M.shape == (self.nb_factors, self.nb_factors)
        mrv = self.meanrev
        mrv2 = mrv * mrv
        mrv3 = mrv * mrv2
        Omega = np.zeros((self.nb_aux_factors,))
        Omega[0] = M[0, 1] / mrv + M[0, 2] / mrv2
        Omega[1] = M[0, 0]
        Omega[2] = -M[0, 1] / mrv - M[0, 2] / mrv2 + M[1, 1] / mrv + M[1, 2] / mrv2
        Omega[3] = M[0, 1] - M[0, 2] / mrv + M[1, 2] / mrv + M[2, 2] / mrv2
        Omega[4] = 2.0 * M[0, 2]
        Omega[5] = -M[1, 1] / mrv - M[1, 2] / mrv2
        Omega[6] = -2.0 / mrv * M[1, 2] - 1.0 / mrv2 * M[2, 2]
        Omega[7] = -2.0 / mrv * M[2, 2]

        return Omega

    def bond(self, t: float, T: float,
               x: np.ndarray, y: np.ndarray,
               ccy: str,
               m: int = 0) -> np.ndarray:
        return self._bond(self.nb_factors, self.nb_aux_factors, t, T, x, y, ccy, m)

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def bond_coeffs(self, tau: float) -> Tuple[np.ndarray, np.ndarray]:
        mrv = self.meanrev
        mrv2 = mrv * mrv
        mrv3 = mrv * mrv * mrv
        mt = mrv * tau
        mt2 = mt * mt
        exp_mrv = np.exp(-mt)
        exp_mrv2 = np.exp(-2.0 * mt)
        B_PX = np.array([tau, (1.0 - exp_mrv) / mrv, (1.0 - exp_mrv * (1.0 + mt)) / mrv2])
        B_PY = np.array([tau, 0.5 * tau * tau,
                         (1.0 - exp_mrv) / mrv, (1.0 - exp_mrv * (1.0 + mt)) / mrv2,
                         (1.0 - exp_mrv * (1.0 + mt + 0.5 * mt2)) / mrv3,
                         0.5 * (1.0 - exp_mrv2) / mrv, 0.25 * (1.0 - exp_mrv2 * (1 + 2.0 * mt))/mrv2,
                         0.125 * (1.0 - exp_mrv2 * (1.0 + 2.0 * mt + 2.0 * mt2)) / mrv3])
        return B_PX, B_PY


#############################################################################
#                     Multi-factor Pure Exponential FHJM model
#############################################################################
@dataclass
class CheyettePEND(BasisHJM):
    mrv0: float
    mrv_delta: float
    key_terms: np.ndarray

    def __post_init__(self):
        assert self.mrv0 > 0
        assert self.mrv_delta > 0
        self.nb_factors = CheyettePEND.get_nb_factors()
        self.nb_aux_factors = CheyettePEND.get_nb_aux_factors()
        # same number of key terms and factors
        assert self.key_terms.size == self.nb_factors

    def calc_mrvs(self):
        mrvs = np.arange(self.mrv0, self.mrv0 + self.mrv_delta * self.nb_factors - 1e-6, self.mrv_delta)
        return mrvs

    def calc_mrvs_extra(self):
        mrvs_extra = np.arange(2.0*self.mrv0, 2.0*self.mrv0 + self.mrv_delta * (2.0 * self.nb_factors - 2.0) + 1e-6, self.mrv_delta)
        return mrvs_extra

    def get_basis(self, tau: float) -> np.ndarray:
        mrvs = self.calc_mrvs()
        exp_mrv = np.exp(-mrvs * tau)
        return exp_mrv

    def get_aux_basis(self, tau: float):
        mrvs = self.calc_mrvs()
        exp_mrvs = np.exp(-mrvs * tau)
        # when lambda_1,...,lambda_d is evenly spaced, then
        # lambda_i + lambda_j is evenly spaced from alpha to alpha + (i+j-2)*delta
        # where alpha = 2*lambda_1
        mrvs_extra = self.calc_mrvs_extra()
        exp_mrvs_extra = np.exp(-mrvs_extra * tau)
        return np.concatenate((exp_mrvs, exp_mrvs_extra))

    @classmethod
    def get_nb_factors(cls) -> int:
        return 3

    @classmethod
    def get_nb_aux_factors(cls) -> int:
        d = cls.get_nb_factors()
        return d + 2*d - 1

    def get_generating_matrix(self) -> np.ndarray:
        D = -np.diag(self.calc_mrvs())
        assert D.shape == (self.nb_factors, self.nb_factors)
        return D

    def get_aux_generating_matrix(self) -> np.ndarray:
        D = -np.diag(np.concatenate((self.calc_mrvs(), self.calc_mrvs_extra())))
        assert D.shape == (self.nb_aux_factors, self.nb_aux_factors)
        return D

    def get_matrix_B(self) -> np.ndarray:
        return self._get_matrix_B(self.nb_factors, self.key_terms)

    def calc_Omega(self, M: np.ndarray) -> np.ndarray:
        assert M.shape == (self.nb_factors, self.nb_factors)
        mrvs = self.calc_mrvs()
        mrvs_extra = self.calc_mrvs_extra()
        assert self.nb_aux_factors == mrvs.size + mrvs_extra.size

        Omega = np.zeros((self.nb_aux_factors,))
        for i in range(mrvs.size):
            Omega[i] = np.dot(M[i,:], 1.0/mrvs)
        for k in range(mrvs_extra.size):
            sum_fix_k = 0.0
            for i,j in zip(range(k,-1,-1), range(0,k+1,1)):
                if 0 <= i < self.nb_factors and 0 <= j < self.nb_factors:
                    sum_fix_k = sum_fix_k - M[i,j]/mrvs[j]
            # assign
            Omega[mrvs.size+k] = sum_fix_k
        return Omega

    def bond(self, t: float, T: float,
             x: np.ndarray, y: np.ndarray,
             ccy: str,
             m: int = 0) -> np.ndarray:
        return self._bond(self.nb_factors, self.nb_aux_factors, t, T, x, y, ccy, m)

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def bond_coeffs(self, tau: float) -> Tuple[np.ndarray, np.ndarray]:
        mrvs = self.calc_mrvs()
        exp_mrvs = np.exp(-mrvs * tau)
        mrvs_extra = self.calc_mrvs_extra()
        exp_mrvs_extra = np.exp(-mrvs_extra * tau)
        B_PX = np.divide(1.0 - exp_mrvs, mrvs)
        B_PY = np.concatenate((np.divide(1.0 - exp_mrvs, mrvs), np.divide(1.0 - exp_mrvs_extra, mrvs_extra)))
        return B_PX, B_PY





"""
key analytics for Black Scholes Merton pricer and implied volatilities
"""

import numpy as np
from numba import njit
from typing import Union
from enum import Enum
from numba.typed import List
from scipy.stats import norm

from stochvolmodels.utils.funcs import ncdf, npdf


class OptionType(str, Enum):
    CALL = 'C'
    PUT = 'P'
    INVERSE_CALL = 'IC'
    INVERSE_PUT = 'IP'


"""
first implement scalar price, delta, vega, gamma along with vectorised versions
nb: numba loops are more efficient than vectorised implementations 
"""


@njit
def is_intrinsic(ttm: float, vol: float) -> bool:
    if ttm <= 0.0 or vol <= 0.0 or np.isnan(vol):
        return True
    else:
        return False


"""
**************************
prices
****************************
"""


@njit
def compute_bsm_vanilla_price(forward: float,
                              strike: float,
                              ttm: float,
                              vol: float,
                              optiontype: str = 'C',
                              discfactor: float = 1.0
                              ) -> float:
    """
    bsm pricer for forward
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        if optiontype == 'C' or optiontype == 'IC':
            price = np.maximum(forward - strike, 0.0)
        elif optiontype == 'P' or optiontype == 'IP':
            price = np.maximum(strike - forward, 0.0)
        else:
            raise NotImplementedError(f"optiontype")

    else:
        s_ttm = vol * np.sqrt(ttm)
        d1 = (np.log(forward / strike) + 0.5 * s_ttm * s_ttm) / s_ttm
        d2 = d1 - s_ttm
        if optiontype == 'C' or optiontype == 'IC':
            price = discfactor * (forward * ncdf(d1) - strike * ncdf(d2))
        elif optiontype == 'P' or optiontype == 'IP':
            price = -discfactor * (forward * ncdf(-d1) - strike * ncdf(-d2))
        else:
            raise NotImplementedError(f"optiontype")
    return price


compute_bsm_vanilla_price_vector = np.vectorize(compute_bsm_vanilla_price, doc='Vectorized `compute_bsm_vanilla_price`')


@njit
def compute_bsm_vanilla_slice_prices(ttm: float,
                                     forward: float,
                                     strikes: np.ndarray,
                                     vols: np.ndarray,
                                     optiontypes: np.ndarray,
                                     discfactor: float = 1.0
                                     ) -> np.ndarray:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    def f(strike: float, vol: float, optiontype: str) -> float:
        return compute_bsm_vanilla_price(forward=forward,
                                         ttm=ttm,
                                         vol=vol,
                                         strike=strike,
                                         optiontype=optiontype,
                                         discfactor=discfactor)

    bsm_prices = np.zeros_like(strikes)
    for idx, (strike, vol, optiontype) in enumerate(zip(strikes, vols, optiontypes)):
        bsm_prices[idx] = f(strike, vol, optiontype)
    return bsm_prices


@njit
def compute_bsm_forward_grid_prices(ttm: float,
                                    forwards: np.ndarray,
                                    strike: float,
                                    vol: float,
                                    optiontype: str,
                                    discfactor: float = 1.0
                                    ) -> np.ndarray:
    """
    vectorised bsm prices for array of aligned forwards
    """
    def f(forward: float) -> float:
        return compute_bsm_vanilla_price(forward=forward,
                                         ttm=ttm,
                                         vol=vol,
                                         strike=strike,
                                         optiontype=optiontype,
                                         discfactor=discfactor)

    bsm_prices = np.zeros_like(forwards)
    for idx, forward in enumerate(forwards):
        bsm_prices[idx] = f(forward)
    return bsm_prices


"""
**************************
deltas
**************************
"""


@njit
def compute_bsm_vanilla_delta(ttm: float,
                              forward: float,
                              strike: float,
                              vol: float,
                              optiontype: str,
                              discfactor: float = 1.0
                              ) -> float:
    """
    bsm deltas for strikes and vols
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        if optiontype == 'C' or optiontype == 'IC':
            bsm_deltas = 1.0 if forward >= strike else 0.0
        elif optiontype == 'P' or optiontype == 'IP':
            bsm_deltas = - 1.0 if forward <= strike else 0.0
        else:
            raise NotImplementedError(f"optiontype")
    else:
        s_ttm = vol * np.sqrt(ttm)
        d1 = np.log(forward / strike) / s_ttm + 0.5 * s_ttm
        if optiontype == 'C':
            d1_sign = 1.0
        elif optiontype == 'P':
            d1_sign = - 1.0
        else:
            d1_sign = 0.0
        bsm_deltas = discfactor * d1_sign * ncdf(d1_sign * d1)
    return bsm_deltas


compute_bsm_vanilla_delta_vector = np.vectorize(compute_bsm_vanilla_delta, doc='Vectorized `compute_bsm_vanilla_delta`')


@njit
def compute_bsm_vanilla_slice_deltas(ttm: float,
                                     forward: float,
                                     strikes: np.ndarray,
                                     vols: np.ndarray,
                                     optiontypes: np.ndarray
                                     ) -> Union[float, np.ndarray]:
    """
    bsm deltas for strikes and vols
    """
    def f(strike: float, vol: float, optiontype: str) -> float:
        return compute_bsm_vanilla_delta(forward=forward,
                                         ttm=ttm,
                                         vol=vol,
                                         strike=strike,
                                         optiontype=optiontype)

    bsm_deltas = np.zeros_like(strikes)
    for idx, (strike, vol, optiontype) in enumerate(zip(strikes, vols, optiontypes)):
        bsm_deltas[idx] = f(strike, vol, optiontype)
    return bsm_deltas


@njit
def compute_bsm_vanilla_deltas_ttms(ttms: np.ndarray,
                                    forwards: np.ndarray,
                                    strikes_ttms: List[np.ndarray],
                                    vols_ttms: List[np.ndarray],
                                    optiontypes_ttms: List[np.ndarray],
                                    ) -> List[np.ndarray]:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    deltas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms, strikes_ttms, optiontypes_ttms):
        deltas_ttms.append(compute_bsm_vanilla_slice_deltas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return deltas_ttms


@njit
def compute_bsm_vanilla_grid_deltas(ttm: float,
                                    forwards: np.ndarray,
                                    strike: float,
                                    vol: float,
                                    optiontype: str,
                                    discfactor: float = 1.0
                                    ) -> np.ndarray:
    """
    vectorised bsm deltas for array of forwards grid
    """
    def f(forward: float) -> float:
        return compute_bsm_vanilla_delta(forward=forward,
                                         ttm=ttm,
                                         vol=vol,
                                         strike=strike,
                                         optiontype=optiontype,
                                         discfactor=discfactor)

    bsm_deltas = np.zeros_like(forwards)
    for idx, forward in enumerate(forwards):
        bsm_deltas[idx] = f(forward)
    return bsm_deltas


def compute_bsm_strike_from_delta(ttm: float,
                                  forward: float,
                                  delta: float,
                                  vol: float
                                  ) -> Union[float, np.ndarray]:
    """
    bsm deltas for strikes and vols
    """
    inv_delta = norm.ppf(delta) if delta > 0.0 else -norm.ppf(-delta)
    s_t = vol * np.sqrt(ttm)
    strike = forward*np.exp(-s_t*(inv_delta - 0.5 * s_t))
    return strike


"""
****************************
Vega
****************************
"""


@njit
def compute_bsm_vanilla_vega(ttm: float,
                             forward: float,
                             strike: float,
                             vol: float,
                             ) -> float:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        vega = 0.0
    else:
        s_t = vol * np.sqrt(ttm)
        d1 = np.log(forward / strike) / s_t + 0.5 * s_t
        vega = forward * npdf(d1) * np.sqrt(ttm)
    return vega


compute_bsm_vanilla_vega_vector = np.vectorize(compute_bsm_vanilla_vega, doc='Vectorized `compute_bsm_vanilla_vega`')


@njit
def compute_bsm_slice_vegas(ttm: float,
                            forward: float,
                            strikes: np.ndarray,
                            vols: np.ndarray,
                            optiontypes: np.ndarray = None
                            ) -> np.ndarray:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    sT = vols * np.sqrt(ttm)
    d1 = np.log(forward / strikes) / sT + 0.5 * sT
    vegas = forward * npdf(d1) * np.sqrt(ttm)
    return vegas


@njit
def compute_bsm_vegas_ttms(ttms: np.ndarray,
                           forwards: np.ndarray,
                           strikes_ttms: List[np.ndarray],
                           vols_ttms: List[np.ndarray],
                           optiontypes_ttms: List[np.ndarray],
                           ) -> List[np.ndarray]:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    vegas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms, strikes_ttms, optiontypes_ttms):
        vegas_ttms.append(compute_bsm_slice_vegas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return vegas_ttms


"""
****************************
Gamma
****************************
"""


@njit
def compute_bsm_vanilla_gamma(ttm: float,
                              forward: float,
                              strike: float,
                              vol: float
                              ) -> float:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        gamma = 0.0
    else:
        s_t = vol * np.sqrt(ttm)
        d1 = np.log(forward / strike) / s_t + 0.5 * s_t
        gamma = npdf(d1) / (forward*s_t)
    return gamma


compute_bsm_vanilla_gamma_vector = np.vectorize(compute_bsm_vanilla_gamma, doc='Vectorized `compute_bsm_vanilla_gamma`')


"""
****************************
Theta
****************************
"""


@njit
def compute_bsm_vanilla_theta(ttm: float,
                              forward: float,
                              strike: float,
                              vol: float,
                              optiontype: str,
                              discfactor: float = 1.0,
                              discount_rate: float = 0.0
                              ) -> float:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        theta = 0.0
    else:
        s_t = vol * np.sqrt(ttm)
        d1 = np.log(forward / strike) / s_t + 0.5 * s_t
        d2 = d1 - s_t
        if optiontype == 'C' or optiontype == 'IC':
            theta = -forward * npdf(d1)*vol/(0.5*np.sqrt(ttm)) - discount_rate*discfactor*strike*ncdf(d2)
        elif optiontype == 'P' or optiontype == 'IP':
            theta = -forward * npdf(d1)*vol/(0.5*np.sqrt(ttm)) + discount_rate*discfactor*strike*ncdf(-d2)
        else:
            raise NotImplementedError(f"optiontype")
    return theta


compute_bsm_vanilla_theta_vector = np.vectorize(compute_bsm_vanilla_theta, doc='Vectorized `compute_bsm_vanilla_theta`')


@njit
def compute_bsm_vanilla_slice_vegas(ttm: float,
                                    forward: float,
                                    strikes: np.ndarray,
                                    vols: np.ndarray,
                                    optiontypes: np.ndarray = None
                                    ) -> np.ndarray:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    s_t = vols * np.sqrt(ttm)
    d1 = np.log(forward / strikes) / s_t + 0.5 * s_t
    vegas = forward * npdf(d1) * np.sqrt(ttm)
    return vegas


@njit
def compute_bsm_vanilla_vegas_ttms(ttms: np.ndarray,
                                   forwards: np.ndarray,
                                   strikes_ttms: List[np.ndarray],
                                   vols_ttms: List[np.ndarray],
                                   optiontypes_ttms: List[np.ndarray],
                                   ) -> List[np.ndarray]:
    """
    vectorised bsm vegas for array of aligned strikes, vols, and optiontypes
    """
    vegas_ttms = List()
    for ttm, forward, vols_ttm, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, vols_ttms,
                                                                    strikes_ttms, optiontypes_ttms):
        vegas_ttms.append(compute_bsm_vanilla_slice_vegas(ttm=ttm, forward=forward, strikes=strikes_ttm, vols=vols_ttm, optiontypes=optiontypes_ttm))
    return vegas_ttms


"""
********************************
implied vols
*******************************
"""


@njit
def infer_bsm_ivols_from_model_slice_prices(ttm: float,
                                            forward: float,
                                            strikes: np.ndarray,
                                            optiontypes: np.ndarray,
                                            model_prices: np.ndarray,
                                            discfactor: float
                                            ) -> np.ndarray:
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_bsm_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                   given_price=model_price,
                                                   strike=strike,
                                                   optiontype=optiontype)
    return model_vol_ttm


@njit
def infer_bsm_implied_vol(forward: float,
                          ttm: float,
                          strike: float,
                          given_price: float,
                          discfactor: float = 1.0,
                          optiontype: str = 'C',
                          tol: float = 1e-16,
                          is_bounds_to_nan: bool = True
                          ) -> float:
    """
    compute Black implied vol
    """
    x1, x2 = 0.01, 5.0  # starting values
    f = compute_bsm_vanilla_price(forward=forward, strike=strike, ttm=ttm, vol=x1, discfactor=discfactor, optiontype=optiontype) - given_price
    fmid = compute_bsm_vanilla_price(forward=forward, strike=strike, ttm=ttm, vol=x2, discfactor=discfactor, optiontype=optiontype) - given_price

    if f*fmid < 0.0:
        if f < 0.0:
            rtb = x1
            dx = x2-x1
        else:
            rtb = x2
            dx = x1-x2
        xmid = rtb
        for j in range(0, 200):
            dx = dx*0.5
            xmid = rtb+dx
            fmid = compute_bsm_vanilla_price(forward=forward, strike=strike, ttm=ttm, vol=xmid, discfactor=discfactor, optiontype=optiontype) - given_price
            if fmid <= 0.0:
                rtb = xmid
            if np.abs(fmid) < tol:
                break
        v1 = xmid

    else:
        if f < 0.0:
            v1 = x1
        else:
            v1 = x2

    if is_bounds_to_nan:  # in case vol was inferred it will return nan
        if np.abs(v1-x1) < tol or np.abs(v1-x2) < tol:
            v1 = np.nan
    return v1


@njit
def infer_bsm_ivols_from_slice_prices(ttm: float,
                                      forward: float,
                                      discfactor: float,
                                      strikes: np.ndarray,
                                      optiontypes: np.ndarray,
                                      model_prices: np.ndarray,
                                      ) -> np.ndarray:
    """
    vectorised chain ivols
    """
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_bsm_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                   given_price=model_price,
                                                   strike=strike,
                                                   optiontype=optiontype)
    return model_vol_ttm


@njit
def infer_bsm_ivols_from_model_chain_prices(ttms: np.ndarray,
                                            forwards: np.ndarray,
                                            discfactors: np.ndarray,
                                            strikes_ttms: List[np.ndarray],
                                            optiontypes_ttms: List[np.ndarray],
                                            model_prices_ttms: List[np.ndarray],
                                            ) -> List[np.ndarray]:
    """
    vectorised chain ivols
    """
    model_vol_ttms = List()
    for ttm, forward, discfactor, strikes, optiontypes, model_prices_ttm in zip(ttms, forwards, discfactors,
                                                                                strikes_ttms, optiontypes_ttms,
                                                                                model_prices_ttms):
        model_vol = np.zeros_like(strikes)
        for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices_ttm, optiontypes)):
            model_vol[idx] = infer_bsm_implied_vol(forward=forward, ttm=ttm, discfactor=discfactor,
                                                   given_price=model_price,
                                                   strike=strike,
                                                   optiontype=optiontype)
        model_vol_ttms.append(model_vol)
    return model_vol_ttms


"""
********************************************
Digital prices
********************************************
"""

@njit
def compute_bsm_digital_price(forward: float,
                              strike: float,
                              ttm: float,
                              vol: float,
                              optiontype: str = 'C',
                              discfactor: float = 1.0
                              ) -> float:
    """
    bsm pricer for forward
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        if optiontype == 'C' or optiontype == 'IC':
            price = 1.0 if forward >= strike else 0.0
        elif optiontype == 'P' or optiontype == 'IP':
            price = 1.0 if forward <= strike else 0.0
        else:
            raise NotImplementedError(f"optiontype")
    else:
        s_ttm = vol * np.sqrt(ttm)
        d1 = (np.log(forward / strike) + 0.5 * s_ttm * s_ttm) / s_ttm
        d2 = d1 - s_ttm
        if optiontype == 'C' or optiontype == 'IC':
            price = discfactor * ncdf(d2)
        elif optiontype == 'P' or optiontype == 'IP':
            price = discfactor * ncdf(-d2)
        else:
            raise NotImplementedError(f"optiontype")

    return price


@njit
def compute_bsm_digital_delta(forward: float,
                              strike: float,
                              ttm: float,
                              vol: float,
                              optiontype: str = 'C',
                              discfactor: float = 1.0
                              ) -> float:
    """
    bsm pricer for forward
    """
    if is_intrinsic(ttm=ttm, vol=vol):
        delta = 0.0
    else:
        s_ttm = vol * np.sqrt(ttm)
        d1 = (np.log(forward / strike) + 0.5 * s_ttm * s_ttm) / s_ttm
        d2 = d1 - s_ttm
        pnorm = discfactor / (forward * s_ttm)
        if optiontype == 'C' or optiontype == 'IC':
            delta = pnorm * npdf(d2)
        elif optiontype == 'P' or optiontype == 'IP':
            delta = - pnorm * npdf(d2)
        else:
            raise NotImplementedError(f"optiontype")

    return delta



import numpy as np
import scipy.integrate as integ
from scipy.optimize import minimize, lsq_linear
from scipy.special import gamma, gammainc

#import orthopy
#import quadpy


def sort(a, b):
    """
    Sorts two numpy arrays jointly according to the ordering of the first.
    :param a: First numpy array
    :param b: Second numpy array
    :return: Sorted numpy arrays
    """
    perm = np.argsort(a)
    return a[perm], b[perm]


def rel_err(x, x_approx):
    """
    Computes the absolute relative error of x_approx compared to x.
    :param x: The true value
    :param x_approx: The approximated value
    :return: The absolute relative error
    """
    return np.abs((x - x_approx) / x)


def single_param_search(f, rel_tol=1e-03, n=100, factor=2):
    """
    Finds the optimal parameter n for approximating f.
    :param f: A function having two inputs n and reusable. The input n is some discretization parameter. For example,
        f might be the solution of an ODE and n is the number of time steps. The input reusable is used to supply
        previously computed information, so that it does not have to be computed again. For example, if f is the
        trapezoidal integral of some function, this may be an array of previously computed function values that can
        be reused without needing to recompute them. If there is nothing sensible that could be reused, just leave it as
        a dead parameter. The function gives two outputs a and b. The output a is the result (e.g. the final point of
        the ODE solution), and the output b is the reusable information. In the next call of f, b will be given as the
        parameter reusable.
    :param rel_tol: Relative error tolerance of the result
    :param n: Initial parameter n
    :param factor: Factor by which we should multiply n if we need higher accuracy.
    :return: The approximated result, the final parameter n that was used, and the reusable information
    """
    int_calc = isinstance(n, int)
    approx_res, reusable = f(n=n // factor if int_calc else n / factor, reusable=None)
    current_res, reusable = f(n=n, reusable=reusable)
    while rel_err(current_res, approx_res) > rel_tol:
        n = int(factor * n) if int_calc else factor * n
        approx_res = current_res
        current_res, reusable = f(n=n, reusable=reusable)
    return current_res, n, reusable


def exp_underflow(x):
    """
    Computes exp(-x) while avoiding underflow errors.
    :param x: Float of numpy array
    :return: exp(-x)
    """
    if isinstance(x, np.ndarray):
        if x.dtype == int:
            x = x.astype(np.float)
        eps = np.finfo(x.dtype).tiny
    else:
        if isinstance(x, int):
            x = float(x)
        eps = np.finfo(x.__class__).tiny
    log_eps = -np.log(eps) / 2
    result = np.exp(-np.fmin(x, log_eps))
    result = np.where(x > log_eps, 0, result)
    return result


def fractional_kernel(H, t):
    """
    The fractional kernel.
    :param H: Hurst parameter
    :param t: Time, may also be a numpy array
    :return: The value of the fractional kernel at t
    """
    return t ** (H - 0.5) / gamma(H + 0.5)


def kernel_norm(H, T, p=2.):
    """
    Returns the L^p-norm of the fractional kernel.
    :param H: Hurst parameter
    :param T: Final time
    :param p: The order of the norm
    :return: The L^p-norm (root has been taken) of the fractional kernel
    """
    return T ** (H - 0.5 + 1 / p) / (gamma(0.5 + H) * (1 + p * H - p / 2) ** (1 / p))


def c_H(H):
    """
    Returns the constant c_H.
    :param H: Hurst parameter
    :return: c_H
    """
    return 1. / (gamma(0.5 + H) * gamma(0.5 - H))


def fractional_kernel_laplace(H, t, nodes):
    """
    The Laplace transform of the fractional kernel.
    :param H: Hurst parameter
    :param t: Time, may also be a numpy array
    :param nodes: Laplace transform argument, may also be a numpy array
    :return: The Laplace transform. May be a number, a one-dimensional or a two-dimensional numpy array depending on
        the shape of t and nodes. If both t and nodes are a numpy array, the tensor product that we take is
        nodes x time
    """
    if isinstance(t, np.ndarray) and isinstance(nodes, np.ndarray):
        return c_H(H) * exp_underflow(np.tensordot(nodes, t, axes=0))
    return c_H(H) * exp_underflow(nodes * t)


def fractional_kernel_approximation(H, t, nodes, weights):
    """
    Returns the Markovian approximation of the fractional kernel.
    :param H: Hurst parameter
    :param t: Time points
    :param nodes: Nodes of the quadrature rule
    :param weights: Weights of the quadrature rule
    :return: The approximated kernel using nodes and weights at times t (a numpy array)
    """
    return 1 / c_H(H) * np.tensordot(fractional_kernel_laplace(H, t, nodes), weights, axes=([0, 0]))


def AK_improved_rule(H, N, K=None, T=1.):
    """
    The quadrature rule from Alfonsi and Kebaier in Table 6, left column.
    :param H: Hurst parameter
    :param N: Total number of nodes
    :param K: Cutoff point where the regime changes
    :param T: Final time
    :return: The quadrature rule in the form nodes, weights
    """
    if N == 1:
        return np.array([0.]), np.array([0.])

    N = N // 2

    if K is None:
        K = N ** 0.8

    def AK_initial_guess(A_):
        partition = np.empty(2 * N + 1)
        partition[:N + 1] = np.linspace(0, K, N + 1)
        partition[N + 1:] = K * A_ ** np.arange(1, N + 1)
        a = partition ** (1.5 - H)
        b = partition ** (0.5 - H)
        nodes_ = (0.5 - H) / (1.5 - H) * (a[1:] - a[:-1]) / (b[1:] - b[:-1])
        weights_ = c_H(H) / (0.5 - H) * (b[1:] - b[:-1])
        return nodes_, weights_

    def error_func(A_):
        nodes_, weights_ = AK_initial_guess(A_[0])
        return error_l2(H, nodes_, weights_, T)

    res = minimize(fun=lambda A_: error_func(A_), x0=np.array([1.2]), bounds=((0, None),))
    A = res.x
    nodes, weights = AK_initial_guess(A[0])

    res = minimize(fun=lambda x: error_l2(H, nodes, x * weights, T), x0=np.array([1]), bounds=((0, None),))
    return nodes, res.x * weights


def AbiJaberElEuch_quadrature_rule(H, N, T):
    """
    Computes the quadrature as suggested in "Multi-factor approximation of rough volatility models" by Abi Jaber and
    El Euch.
    :param H: Hurst parameter
    :param N: Number of quadrature nodes
    :param T: Maturity / Final time
    :return: The nodes and weights, two numpy arrays
    """
    pi_n = N ** (-0.2) / T * (np.sqrt(10) * (1 - 2 * H) / (5 - 2 * H)) ** 0.4
    eta = pi_n * np.arange(N + 1)
    c_vec = (eta[1:] ** (0.5 - H) - eta[:-1] ** (0.5 - H)) / (gamma(H + 0.5) * gamma(1.5 - H))
    gamma_vec = (eta[1:] ** (1.5 - H) - eta[:-1] ** (1.5 - H)) / ((1.5 - H) * gamma(H + 0.5) + gamma(0.5 - H)) / c_vec
    return gamma_vec, c_vec


def Gaussian_parameters(H, N, T, mode):
    """
    Returns the parameters of the Gaussian quadrature rule.
    :param H: Hurst parameter
    :param N: Total number of nodes
    :param T: Final time
    :param mode: The kind of theorem or observation that should be used
    :return: The partition of the middle part, and the quadrature level m
    """
    if ' geometric ' in mode or mode == "OLD" or mode == "GG":
        if mode == "old geometric theorem l2":
            N = N - 1
            A = np.sqrt(1 / H + 1 / (1.5 - H))
            beta = 0.4275
            alpha = 1.06418
            gamma_ = np.exp(alpha * beta)
            exponent = 1 / (3 * gamma_ / (8 * (gamma_ - 1)) + 6 * H - 4 * H * H)
            temp_1 = ((9 - 6 * H) / (2 * H)) ** (gamma_ / (8 * (gamma_ - 1)))
            temp_2 = 5 * np.pi ** 3 * gamma_ * (gamma_ - 1) * A ** (2 - 2 * H) * float(N) ** (1 - H) / (
                        beta ** (2 - 2 * H))
            base_0 = temp_1 * (temp_2 * (3 - 2 * H) / (768 * H)) ** (2 * H)
            a = 1 / T * base_0 ** exponent * np.exp(-alpha / ((1.5 - H) * A) * np.sqrt(N))
            base_n = temp_1 * (temp_2 / 1152) ** (2 * H - 3)
            b = 1 / T * base_n ** exponent * np.exp(alpha / (H * A) * np.sqrt(N))
            m = int(np.fmax(np.round(beta / A * np.sqrt(N)), 1))
            n = int(np.round(N / m))
        elif mode == "old geometric observation l2" or mode == "OLD":
            N = N - 1
            A = np.sqrt(1 / H + 1 / (1.5 - H))
            beta = 0.9
            alpha = 1.8
            a = 0.65 * 1 / T * np.exp(3.1 * H) * np.exp(-alpha / ((1.5 - H) * A) * np.sqrt(N))
            b = 1 / T * np.exp(3 * H ** (-0.4)) * np.exp(alpha / (H * A) * np.sqrt(N))
            m = int(np.fmax(np.round(beta / A * np.sqrt(N)), 1))
            n = int(np.round(N / m))
        elif mode == "new geometric theorem l1" or mode == "GG":
            beta = 1
            alpha = np.log(3 + 2 * np.sqrt(2))
            a = 4 / T
            b = 1 / 2 / T * np.exp(alpha / np.sqrt(H + 0.5) * np.sqrt(N))
            m = int(np.fmax(np.round(beta * np.sqrt((H + 0.5) * N)), 1))
            n = int(np.round(N / m)) - 1
        else:
            raise NotImplementedError(f'The mode {mode} has not been implemented')

        partition = np.exp(np.log(a) + np.log(b / a) * np.linspace(0, 1, n + 1))
    else:
        if mode == 'non-geometric l1' or mode == "NGG":
            beta = 0.92993273
            a = 3 / T
            m = int(np.fmax(np.round(beta * np.sqrt((H + 0.5) * N)), 1))
            c = 3.60585021
        else:
            raise NotImplementedError(f'The mode {mode} has not been implemented')

        kappa = 1 / (2 * beta ** 2)
        n = int(np.round(N / m)) - 1
        partition = np.empty(n + 1)
        partition[0] = a
        for i in range(n):
            partition[i + 1] = partition[i] \
                * ((c + partition[i] ** (kappa / (n + 1))) / (c - partition[i] ** (kappa / (n + 1)))) ** 2
    return partition, m


def Gaussian_interval(H, m, a, b, fractional_weight=True):
    """
    Returns the nodes and weights of the Gauss quadrature rule of level m on [a, b].
    :param H: Hurst parameter
    :param m: Level of the Gaussian quadrature
    :param a: Left end of interval
    :param b: Right end of interval
    :param fractional_weight: If True, computes the Gaussian quadrature rule with respect to the fractional weight. If
        False, computes the Gaussian quadrature with respect to the weight function w(x) = c_H
    :return: The nodes and weights
    """
    if fractional_weight:
        k = np.arange(2 * m) + 0.5 - H
    else:
        k = np.arange(1, 2 * m + 1)
    alpha, beta, int_1 = orthopy.tools.chebyshev(moments=c_H(H) / k * (b ** k - a ** k))
    return quadpy.tools.scheme_from_rc(alpha, beta, int_1)


def Gaussian_on_partition(H, m, partition, fractional_weight=True):
    """
    Returns the nodes and weights of the Gaussian quadrature rule of level m on a partition.
    :param H: Hurst parameter
    :param m: Level of the quadrature rule
    :param partition: The partition, where the Gaussian quadrature rule is applied on each interval
    :param fractional_weight: If True, computes the Gaussian quadrature rule with respect to the fractional weight
        function. If False, computes the Gaussian quadrature with respect to the weight function w(x) = c_H, and
        then multiplies the weights by nodes ** (-H - 1/2)
    :return: All the nodes and weights
    """
    nodes = np.empty(m * (len(partition) - 1))
    weights = np.empty(m * (len(partition) - 1))
    for i in range(len(partition) - 1):
        new_nodes, new_weights = Gaussian_interval(H=H, m=m, a=partition[i], b=partition[i + 1],
                                                   fractional_weight=fractional_weight)
        nodes[m * i:m * (i + 1)] = new_nodes
        weights[m * i:m * (i + 1)] = new_weights
    if not fractional_weight:
        weights = weights * nodes ** (-H - 0.5)
    return nodes, weights


def Gaussian_optimal_zero_weight(H, T, nodes, weights):
    """
    Computes the optimal weight in the L^2-approximation of an additional node at 0 given that we are already using the
    specified nodes and weights.
    :param H: Hurst parameter
    :param T: Final time
    :param nodes: The nodes of the Markovian approximation, a numpy array
    :param weights: The weights of the Markovian approximation, a numpy array
    :return: The optimal weight in the L^2-sense of an additional node at 0
    """
    if len(nodes) == 0:
        return T ** (H - 0.5) / gamma(H + 1.5)
    return (T ** (H + 0.5) / gamma(H + 1.5) - np.sum(weights / nodes * (1 - exp_underflow(nodes * T)))) / T


def Gaussian_rule(H, N, T, mode):
    """
    Returns the nodes and weights of the Gaussian rule with roughly N nodes.
    :param H: Hurst parameter
    :param N: Number of nodes
    :param T: Final time
    :param mode: The Gaussian parameters that should be used
    :return: The nodes and weights, ordered by the size of the nodes
    """
    if isinstance(T, np.ndarray):
        T = T[-1]
    partition, m = Gaussian_parameters(H, N, T, mode)

    if mode == 'old geometric theorem l2' or mode == 'old geometric observation l2':
        if N == 1:
            w_0 = Gaussian_optimal_zero_weight(H=H, T=T, nodes=np.array([]), weights=np.array([]))
            nodes, weights = np.array([0.]), np.array([w_0])
        else:
            nodes, weights = np.zeros(m * (len(partition) - 1) + 1), np.empty(m * (len(partition) - 1) + 1)
            nodes[1:], weights[1:] = Gaussian_on_partition(H=H, m=m, partition=partition, fractional_weight=True)
            weights[0] = Gaussian_optimal_zero_weight(H=H, T=T, nodes=nodes[1:], weights=weights[1:])
    else:
        nodes, weights = np.empty(m * len(partition)), np.empty(m * len(partition))
        nodes[:m], weights[:m] = Gaussian_interval(H=H, m=m, a=0, b=partition[0], fractional_weight=True)
        if len(partition) > 1:
            nodes[m:], weights[m:] = Gaussian_on_partition(H=H, m=m, partition=partition,
                                                           fractional_weight='old' in mode)
    return nodes, weights


def error_l2(H, nodes, weights, T, output='error'):
    """
    Computes an error estimate of the squared L^2-norm of the difference between the rough kernel and its approximation
    on [0, T].
    :param H: Hurst parameter
    :param nodes: The nodes of the approximation. Assumed that they are all non-zero
    :param weights: The weights of the approximation
    :param T: Final time, may also be a numpy array
    :param output: If error, returns the error. If gradient, returns the error and the gradient of the error
    :return: An error estimate
    """
    nodes = np.fmin(np.fmax(nodes, 1e-08), 1e+150)
    weights = np.fmin(weights, 1e+75)
    weight_matrix = np.outer(weights, weights)
    summand = T ** (2 * H) / (2 * H * gamma(H + 0.5) ** 2)
    node_matrix = nodes[:, None] + nodes[None, :]
    if isinstance(T, np.ndarray):
        gamma_ints = gammainc(H + 0.5, np.outer(T, nodes))
        nmT = np.einsum('i,jk->ijk', T, node_matrix)
        exp_node_matrix = exp_underflow(nmT)
        sum_1 = np.sum((weight_matrix / node_matrix)[None, :, :] * (1 - exp_node_matrix), axis=(1, 2))
        sum_2 = 2 * np.sum((weights / nodes ** (H + 0.5))[None, :] * gamma_ints, axis=1)
    else:
        gamma_ints = gammainc(H + 0.5, nodes * T)
        nmT = node_matrix * T
        exp_node_matrix = exp_underflow(nmT)
        sum_1 = np.sum(weight_matrix / node_matrix * (1 - exp_node_matrix))
        sum_2 = 2 * np.sum(weights / nodes ** (H + 0.5) * gamma_ints)
    err = summand + sum_1 - sum_2
    if output == 'error' or output == 'err':
        return err

    N = len(nodes)
    if isinstance(T, np.ndarray):
        grad = np.empty((len(T), 2 * N))
        exp_node_vec = exp_underflow(np.outer(T, nodes)) / nodes[None, :]
        first_summands = (weight_matrix / (node_matrix * node_matrix))[None, :] * (1 - (1 + nmT) * exp_node_matrix)
        second_summands = weights[None, :] * ((T ** (H + 1 / 2) / gamma(H + 1 / 2))[:, None] * exp_node_vec - (
                ((H + 1 / 2) * nodes ** (-H - 3 / 2))[None, :] * gamma_ints))
        grad[:, :N] = -2 * np.sum(first_summands, axis=2) - 2 * second_summands
        third_summands = np.einsum('ijk,k->ij', ((1 - exp_node_matrix) / node_matrix[None, :, :]), weights)
        forth_summands = (nodes ** (-(H + 1 / 2)))[None, :] * gamma_ints
        grad[:, N:] = 2 * third_summands - 2 * forth_summands
    else:
        grad = np.empty(2 * N)
        exp_node_vec = np.zeros(N)
        indices = nodes * T < 300
        exp_node_vec[indices] = np.exp(- T * nodes[indices]) / nodes[indices]
        first_summands = weight_matrix / (node_matrix * node_matrix) * (1 - (1 + nmT) * exp_node_matrix)
        second_summands = weights * (T ** (H + 1 / 2) / gamma(H + 1 / 2) * exp_node_vec - (H + 1 / 2) * nodes ** (
                -H - 3 / 2) * gamma_ints)
        grad[:N] = -2 * np.sum(first_summands, axis=1) - 2 * second_summands
        third_summands = ((1 - exp_node_matrix) / node_matrix) @ weights
        forth_summands = nodes ** (-(H + 1 / 2)) * gamma_ints
        grad[N:] = 2 * third_summands - 2 * forth_summands
    return err, grad


def error_l1(H, nodes, weights, T, method='intersections', tol=1e-08):
    """
    Computes an error estimate of the L^1-norm of the difference between the rough kernel and its approximation
    on [0, T].
    :param H: Hurst parameter
    :param nodes: The nodes of the approximation. Assumed that they are all non-zero
    :param weights: The weights of the approximation
    :param T: Final time, may also be a numpy array
    :param method: Method for computing the error
    :param tol: Relative error tolerance with which the error should be computed (relative error of the error)
    :return: An error estimate
    """
    if method == 'trapezoidal':
        def error_(n, reusable):
            t = np.linspace(0, T, n + 1)[1:]
            if reusable is None:
                reusable = np.abs(fractional_kernel(H, t) - fractional_kernel_approximation(H, t, nodes, weights))
            else:
                error_t_1 = np.empty(n)
                error_t_1[1::2] = reusable
                error_t_1[::2] = np.abs(fractional_kernel(H, t[1::2])
                                        - fractional_kernel_approximation(H, t[1::2], nodes, weights))
                reusable = error_t_1
            total_error = np.trapz(reusable, dx=T / n)
            return total_error + np.abs(fractional_kernel(H, T / (2 * n))
                                        - fractional_kernel_approximation(H, T / (2 * n), nodes, weights)) * T / n, \
                reusable
        return single_param_search(f=error_, rel_tol=tol, n=100, factor=2)[0:2]
    elif method == 'exact - trapezoidal':
        gamma_ = gamma(H + 0.5)

        def find_first_intersection():
            current_error_ = 10.
            current_t = 0.
            current_kernel_approximation = fractional_kernel_approximation(H=H, t=current_t, nodes=nodes,
                                                                           weights=weights)
            while current_error_ > tol and current_t < T:
                current_t = (current_kernel_approximation * gamma_) ** (1 / (H - 0.5))
                current_kernel_approximation = fractional_kernel_approximation(H=H, t=current_t, nodes=nodes,
                                                                               weights=weights)
                current_kernel = fractional_kernel(H=H, t=current_t)
                current_error_ = rel_err(current_kernel, current_kernel_approximation)
            return np.fmin(current_t, T)

        def error_(n, reusable):
            t = np.linspace(t_0, T, n + 1)
            if reusable is None:
                reusable = np.abs(fractional_kernel(H, t) - fractional_kernel_approximation(H, t, nodes, weights))
            else:
                error_t_1 = np.empty(n + 1)
                error_t_1[::2] = reusable
                error_t_1[1::2] = np.abs(fractional_kernel(H, t[1::2])
                                         - fractional_kernel_approximation(H, t[1::2], nodes, weights))
                reusable = error_t_1
            total_error = np.trapz(reusable, dx=(T - t_0) / n)
            return error_to_t_0 + total_error, reusable

        t_0 = find_first_intersection()
        error_to_t_0 = t_0 ** (H + 0.5) / (gamma_ * (H + 0.5)) \
            - np.sum(weights / nodes * (1 - exp_underflow(nodes * t_0)))
        if t_0 == T:
            return error_to_t_0
        return single_param_search(f=error_, rel_tol=tol, n=100, factor=2)[0:2]
    elif method == 'upper bound':
        gamma_ = gamma(H + 0.5)
        nm = nodes[:, None] + nodes[None, :]
        wm = weights[:, None] * weights[None, :]
        err = - 2 * gamma_ * gamma(1.5 + H) * np.sum(weights / nodes ** (1.5 + H) * gammainc(1.5 + H, nodes * T))\
            + gamma_ ** 2 * gamma(2) * np.sum(wm / nm ** 2 * gammainc(2, nm * T))
        return err, 0
    elif method == 'intersections':
        gamma_1 = gamma(H + 0.5)

        def step(t_, ker_, ker_approx_):
            nonlocal n_steps
            n_steps = n_steps + 1
            ker_larger = ker_ > ker_approx_
            rel_err_ = rel_err(ker_, ker_approx_)
            if rel_err_ > tol:
                d_ker = (H - 0.5) / gamma_1 * t_ ** (H - 1.5)
                d_ker_approx = - np.sum(weights * nodes * exp_underflow(nodes * t_))
                if ker_larger:
                    dd_ker_approx = np.sum(weights * nodes ** 2 * exp_underflow(nodes * t_))
                    return t_ + (d_ker - d_ker_approx
                                 + np.sqrt((d_ker - d_ker_approx) ** 2 - 2 * dd_ker_approx * (ker_approx_ - ker_))) \
                        / dd_ker_approx
                else:
                    dd_ker = (H - 0.5) * (H - 1.5) / gamma_1 * t_ ** (H - 2.5)
                    return t_ + (d_ker_approx - d_ker + np.sqrt((d_ker - d_ker_approx) ** 2
                                                                - 2 * dd_ker * (ker_ - ker_approx_))) / dd_ker
            else:
                t_1 = t_ + tol * ker_ / np.sum(weights * nodes * exp_underflow(nodes * t_))
                t_2 = t_ + tol * ker_ ** 2 / (tol * ker_ + ker_approx_) * gamma_1 / (0.5 - H) * t_ ** (1.5 - H)
                return np.fmin(t_1, t_2)

        def find_next_intersection(t_):
            ker_ = fractional_kernel(H=H, t=t_)
            ker_approx_ = fractional_kernel_approximation(H=H, t=t_, nodes=nodes, weights=weights)
            ker_larger = ker_ > ker_approx_
            t_old = t_
            while (ker_ > ker_approx_) == ker_larger and t_ < T:
                t_old = t_
                t_ = step(t_=t_, ker_=ker_, ker_approx_=ker_approx_)
                ker_ = fractional_kernel(H=H, t=t_)
                ker_approx_ = fractional_kernel_approximation(H=H, t=t_, nodes=nodes, weights=weights)
            return t_old, np.fmin(t_, T)

        n_steps = 0
        err = 0
        ker_approx = fractional_kernel_approximation(H=H, t=0, nodes=nodes, weights=weights)
        last_t = 0
        while last_t < T:
            t_left, t_right = find_next_intersection(t_=(ker_approx * gamma_1) ** (1 / (H - 0.5))
                                                     if last_t == 0 else last_t)
            this_t = (t_left + t_right) / 2 if t_right < T else T
            err = err + np.abs((this_t ** (H + 0.5) - last_t ** (H + 0.5)) / (gamma_1 * (H + 0.5))
                               - np.sum(weights / nodes * exp_underflow(last_t * nodes)
                                        * (1 - exp_underflow((this_t - last_t) * nodes))))
            last_t = this_t
        return err, n_steps
    elif method == 'reparametrized trapezoidal':

        def error_(n, reusable):
            t = np.linspace(0, T ** (0.5 + H), n + 1)[1:] ** (1 / (0.5 + H))
            if reusable is None:
                reusable = np.empty(n + 1)
                reusable[1:] = np.abs(1 / gamma(H + 1.5)
                                      - t ** (0.5 - H) * fractional_kernel_approximation(H, t, nodes,
                                                                                         weights) / (0.5 + H))
                reusable[0] = 1 / gamma(H * 1.5)
            else:
                error_t_1 = np.empty(n + 1)
                error_t_1[::2] = reusable
                error_t_1[1::2] = np.abs(1 / gamma(H + 1.5)
                                         - t[1::2] ** (0.5 - H) * fractional_kernel_approximation(H, t[1::2], nodes,
                                                                                                  weights) / (0.5 + H))
                reusable = error_t_1
            total_error = np.trapz(reusable, dx=T ** (0.5 + H) / n)
            return total_error, reusable

        return single_param_search(f=error_, rel_tol=tol, n=100, factor=2)[0:2]
    elif method == 'gaussian':
        return kernel_norm(H=H, T=T, p=1.) - np.sum(weights / nodes * (1 - exp_underflow(nodes * T)))
    else:
        raise NotImplementedError(f'The method {method} for computing the L^1 kernel error has not been implemented.')


def error_l2_optimal_weights(H, T, nodes, output='error'):
    """
    Computes an error estimate of the squared L^2-norm of the difference between the rough kernel and its approximation
    on [0, T]. Uses the best possible weights given the nodes specified.
    :param H: Hurst parameter
    :param nodes: The nodes of the approximation. Assumed that they are all non-zero
    :param output: If error, returns the error and the optimal weights. If gradient, returns the error, the gradient
        (of the nodes only), and the optimal weights. If hessian, returns the error, the gradient, the Hessian, and
        the optimal weights
    :param T: Final time, may also be a numpy array
    :return: An error estimate
    """
    if len(nodes) == 1:
        node = np.fmax(1e-04, nodes[0])
        gamma_1 = gamma(H + 0.5)

        if isinstance(T, np.ndarray):
            nT = node * T
            gamma_ints = gammainc(H + 0.5, nT)
            exp_node_matrix = exp_underflow(2 * nT)
            exp_node_vec = exp_underflow(nT)
            A = (1 - exp_node_matrix) / (2 * node)
            b = -2 * gamma_ints / node ** (H + 0.5)
            c = T ** (2 * H) / (2 * H * gamma_1 ** 2)
            v = b / A
            err = c - 0.25 * b * v
            opt_weights = -0.5 * v
            if len(opt_weights.shape) > 1:
                opt_weights = opt_weights[-1, ...]
            if output == 'error' or output == 'err':
                return err, opt_weights

            A_grad = (-1 + (1 + 2 * nT) * exp_node_matrix) / (2 * node) ** 2
            b_grad = -2 * (nT ** (H + 0.5) * exp_node_vec[None, :] / gamma_1 - (H + 0.5) * gamma_ints) \
                / node ** (H + 1.5)
            grad = 0.5 * A_grad * v ** 2 - 0.5 * b_grad * v
            if output == 'gradient' or output == 'grad':
                return err, grad, opt_weights

            A_hess = 2 * (1 - (1 + 2 * nT + 2 * nT ** 2) * exp_node_matrix) / (8 * node ** 3)
            b_hess = -2 * (-(nT ** (H + 1.5) + (H + 1.5) * nT ** (H + 0.5)) * exp_node_vec / gamma_1 + (H + 0.5) * (
                    H + 1.5) * gamma_ints) / nodes ** (H + 2.5)
            U = b_grad / A
            Y = 2 * A_grad * v
            hess = 0.5 * (2 * Y * U - Y ** 2 / A + 2 * A_hess * v ** 2 - b_hess * v - b_grad * U)
            return err, grad, hess, opt_weights

        gamma_ints = gammainc(H + 0.5, node * T)
        exp_node_matrix = exp_underflow(2 * node * T)
        exp_node_vec = exp_underflow(node * T)
        A = (1 - exp_node_matrix) / (2 * node)
        b = -2 * gamma_ints / node ** (H + 0.5)
        if H > 0:
            c = T ** (2 * H) / (2 * H * gamma_1 ** 2)
            v = b / A
            err = c - 0.25 * b * v
            opt_weight = np.array([-0.5 * v])
            if output == 'error' or output == 'err':
                return err, opt_weight
        else:
            v = b / A
            err = - 0.25 * b * v
            opt_weight = np.array([-0.5 * v])
            if output == 'error' or output == 'err':
                return err, opt_weight

        A_grad = (-1 + (1 + 2 * node * T) * exp_node_matrix) / (4 * node ** 2)
        b_grad = -2 * ((node * T) ** (H + 0.5) * exp_node_vec / gamma_1 - (H + 0.5) * gamma_ints) / node ** (H + 1.5)
        grad = 0.5 * (A_grad * v - b_grad) * v
        if output == 'gradient' or output == 'grad':
            return err, grad, opt_weight

        A_hess = 2 * (1 - (1 + 2 * node * T + 2 * (node * T) ** 2) * exp_node_matrix) / (8 * node ** 3)
        b_hess = -2 * (-((node * T) ** (H + 1.5) + (H + 1.5) * (node * T) ** (H + 0.5)) * exp_node_vec / gamma_1
                       + (H + 0.5) * (H + 1.5) * gamma_ints) / node ** (H + 2.5)
        U = b_grad / A
        Y = 2 * A_grad * v
        hess = 0.5 * (2 * Y * U - Y ** 2 / A + 2 * A_hess * v ** 2 - b_hess * v - b_grad * U)
        return err, grad, hess, opt_weight

    def invert_permutation(p):
        s = np.empty_like(p)
        s[p] = np.arange(p.size)
        return s

    perm = np.argsort(nodes)
    nodes = nodes[perm]
    nodes[0] = np.fmax(1e-04, nodes[0])
    for i in range(len(nodes) - 1):
        if 1.01 * nodes[i] > nodes[i + 1]:
            nodes[i + 1] = nodes[i] * 1.01
    nodes = nodes[invert_permutation(perm)]

    node_matrix = nodes[:, None] + nodes[None, :]
    gamma_1 = gamma(H + 0.5)

    if isinstance(T, np.ndarray):
        nT = np.outer(T, nodes)
        nmT = np.einsum('i,jk->ijk', T, node_matrix)
        gamma_ints = gammainc(H + 0.5, nT)
        exp_node_matrix = exp_underflow(nmT)
        exp_node_vec = exp_underflow(nT)
        A = (1 - exp_node_matrix) / node_matrix[None, :, :]
        b = -2 * gamma_ints / nodes[None, :] ** (H + 0.5)
        c = T ** (2 * H) / (2 * H * gamma_1 ** 2)
        try:
            v = np.linalg.solve(A, b)
        except np.linalg.LinAlgError:
            v = np.empty((len(T), len(nodes)))
            for i in range(len(T)):
                try:
                    v[i, :] = np.linalg.solve(A[i, ...], b[i, ...])
                except np.linalg.LinAlgError:
                    v[i, :] = np.linalg.lstsq(A[i, ...], b[i, ...], rcond=None)[0]
        err = c - 0.25 * np.sum(b * v, axis=1)
        opt_weights = -0.5 * v
        if len(opt_weights.shape) > 1:
            opt_weights = opt_weights[-1, ...]
        if output == 'error' or output == 'err':
            return err, opt_weights

        def mvp(A_, b_):
            return np.sum(A_ * b_[:, None, :], axis=-1)

        A_grad = (-1 + (1 + nmT) * exp_node_matrix[None, :, :]) / node_matrix[None, :, :] ** 2
        b_grad = -2 * (nT ** (H + 0.5) * exp_node_vec[None, :] / gamma_1 - (H + 0.5) * gamma_ints) \
            / nodes[None, :] ** (H + 1.5)
        grad = 0.5 * v * mvp(A_grad, v) - 0.5 * b_grad * v
        if output == 'gradient' or output == 'grad':
            return err, grad, opt_weights

        def diagonalize(x):
            new_x = np.empty((x.shape[0], x.shape[1], x.shape[1]))
            for k in range(x.shape[0]):
                new_x[k, :, :] = np.diag(x[k, :])
            return new_x

        def trans(x):
            return np.transpose(x, (0, 2, 1))

        A_hess = 2 * (1 - (1 + nmT + nmT ** 2 / 2) * exp_node_matrix[None, :, :]) / node_matrix[None, :, :] ** 3
        b_hess = -2 * (-(nT ** (H + 1.5) + (H + 1.5) * nT ** (H + 0.5)) * exp_node_vec / gamma_1 + (H + 0.5) * (
                    H + 1.5) * gamma_ints) / nodes[None, :] ** (H + 2.5)
        try:
            U = np.linalg.solve(A, diagonalize(b_grad))
        except np.linalg.LinAlgError:
            diag_b = diagonalize(b_grad)
            U = np.empty((len(T), len(nodes), len(nodes)))
            for i in range(len(T)):
                for j in range(len(nodes)):
                    try:
                        U[i, j, :] = np.linalg.solve(A[i, ...], diag_b[i, j, :])
                    except np.linalg.LinAlgError:
                        U[i, j, :] = np.linalg.lstsq(A[i, ...], diag_b[i, j, :])[0]
        Y = diagonalize(mvp(A_grad, v)) + A_grad * v[:, None, :]
        YTU = trans(Y) @ U
        hess = 0.5 * (YTU - trans(np.linalg.solve(A, Y)) @ Y + diagonalize(v * mvp(A_hess, v))
                      + v[:, None, :] * v[:, :, None] * A_hess - diagonalize(b_hess * v) - b_grad[:, :, None] * U
                      + trans(YTU))
        return err, grad, hess, opt_weights

    nT = nodes * T
    nmT = node_matrix * T
    gamma_ints = gammainc(H + 0.5, nT)
    exp_node_matrix = exp_underflow(nmT)
    A = (1 - exp_node_matrix) / node_matrix
    b = -2 * gamma_ints / nodes ** (H + 0.5)
    c = T ** (2 * H) / (2 * H * gamma_1 ** 2)
    try:
        v = np.linalg.solve(A, b)
    except np.linalg.LinAlgError:
        v = np.linalg.lstsq(A, b, rcond=None)[0]
    if np.amax(v) > 0:
        v = lsq_linear(A, b).x
    err = 0.25 * v @ A @ v - 0.5 * np.dot(b, v) + c  # c - 0.25 * np.dot(b, v)
    opt_weights = -0.5 * v
    if output == 'error' or output == 'err':
        return err, opt_weights

    exp_node_vec = exp_underflow(nT)
    A_grad = (-1 + (1 + nmT) * exp_node_matrix) / node_matrix ** 2
    b_grad = -2 * (nT ** (H + 0.5) * exp_node_vec / gamma_1 - (H + 0.5) * gamma_ints) / nodes ** (H + 1.5)
    grad = 0.5 * v * (A_grad @ v) - 0.5 * b_grad * v
    if output == 'gradient' or output == 'grad':
        return err, grad, opt_weights

    A_hess = 2 * (1 - (1 + nmT + nmT ** 2 / 2) * exp_node_matrix) / node_matrix ** 3
    b_hess = -2 * (-(nT ** (H + 1.5) + (H + 1.5) * nT ** (H + 0.5)) * exp_node_vec / gamma_1 + (H + 0.5) * (
            H + 1.5) * gamma_ints) / nodes ** (H + 2.5)
    try:
        U = np.linalg.solve(A, np.diag(b_grad))
    except np.linalg.LinAlgError:
        U = np.linalg.lstsq(A, b, rcond=None)[0]
    Y = np.diag(A_grad @ v) + A_grad * v[None, :]
    YTU = Y.T @ U
    hess = 0.5 * (YTU - np.linalg.solve(A, Y).T @ Y + np.diag(v * (A_hess @ v)) + v[None, :] * v[:, None] * A_hess
                  - np.diag(b_hess * v) - b_grad[:, None] * U + YTU.T)
    return err, grad, hess, opt_weights


def optimize_error_l2(H, N, T, tol=1e-08, bound=None, method='gradient', force_order=False, init_nodes=None,
                      iterative=False):
    """
    Optimizes the L^2 error with N points for the fractional kernel. Always uses the best weights and only numerically
    optimizes over the nodes.
    :param H: Hurst parameter
    :param N: Number of points
    :param T: Final time, may be a numpy array (only if grad is False and fast is True)
    :param tol: Error tolerance
    :param bound: Upper bound on the nodes. If no upper bound is desired, use None
    :param method: If error, uses only the error estimates for optimizing over the nodes, and uses the optimizer
        L-BFGS-B. If gradient, uses also the gradient of the error with respect to the nodes, and uses the optimizer
        L-BFGS-B. If hessian, uses also the gradient and the Hessian of the error with respect to the nodes, and uses
        the optimizer trust-constr
    :param force_order: Forces the nodes to stay in order, i.e. not switch places. May improve numerical stability
    :param init_nodes: May specify a starting point for the nodes
    :param iterative: If True, starts with 1 node and iteratively solves the optimization problem before adding another
        node
    :return: The minimal error together with the associated nodes and weights.
    """
    error_fun = error_l2_optimal_weights
    all_errors = np.empty(1)

    if iterative and not init_nodes and N >= 2:
        all_errors = np.empty(N)
        init_nodes = np.empty(N)
        all_errors[:-1], init_nodes[:-1], _ = optimize_error_l2(H=H, N=N - 1, T=T, tol=tol, bound=bound, method=method,
                                                                force_order=force_order, init_nodes=None,
                                                                iterative=iterative)
        init_nodes[:-1] = init_nodes[:N - 1] / 1.03 ** np.fmin(np.arange(1, N) ** 2, 100)
        if bound is not None:
            init_nodes[N - 1] = np.fmax(bound, 10 * init_nodes[N - 2])
        else:
            init_nodes[N - 1] = 5 * init_nodes[N - 2]

    # get starting value and bounds for the optimization problem
    if init_nodes is None:
        if bound is None:
            bound = 1e+100
            nodes_, w = quadrature_rule(H, N, T, mode='old geometric observation l2')
            if N == 2:
                bound = np.fmax(bound, np.amax(nodes_))
            if len(nodes_) < N:
                nodes = np.zeros(N)
                nodes[:len(nodes_)] = nodes_
                for i in range(len(nodes_), N):
                    nodes[i] = nodes_[-1] * 10 ** (i - len(nodes_) + 1)
            else:
                nodes = nodes_[:N]
        else:
            nodes = np.exp(np.linspace(0, np.log(np.fmin(bound, 5. ** (np.fmin(140, N - 1)) / T)), N))
    else:
        if bound is None:
            bound = 1e+100
        nodes = init_nodes
    lower_bound = 1 / (10 * N * np.amin(T)) * ((0.5 - H) / 0.4) ** 2
    nodes = np.fmin(np.fmax(nodes, lower_bound), bound)
    bounds = ((np.log(lower_bound), np.log(bound)),) * N
    original_error, original_weights = error_fun(H=H, T=T, nodes=nodes, output='error')
    original_nodes = nodes.copy()

    # carry out the optimization
    if force_order:
        constraints = []
        for i in range(1, N):
            def jac_here(x):
                res_ = np.zeros(N)
                res_[i] = 1
                res_[i - 1] = -1
                return res_
            constraints = constraints + [{'type': 'ineq', 'fun': lambda x: x[i] - x[i - 1] - 0.3, 'jac': jac_here}]

        if method == 'error' or method == 'err':
            def func(x):
                return error_fun(H, T, np.exp(x), output='error')[0] \
                    * (1 + np.sum(np.exp(- (x[1:] - x[:-1]) ** 2 * 3 * N / (0.72 * np.log(5 / H) - np.log(T)))))

            res = minimize(func, np.log(nodes), tol=tol ** 2, bounds=bounds, constraints=constraints)

        else:
            def func(x):
                err_, grad, _ = error_fun(H, T, np.exp(x), output='gradient')
                return err_, np.exp(x) * grad

            res = minimize(func, np.log(nodes), tol=tol ** 2, bounds=bounds, constraints=constraints, jac=True)

    else:
        if method == 'error' or method == 'err':
            def func(x):
                return error_fun(H, T, np.exp(x), output='error')[0]

            res = minimize(func, np.log(nodes), tol=tol ** 2, bounds=bounds)

        elif method == 'gradient' or method == 'grad':
            def func(x):
                err_, grad, _ = error_fun(H, T, np.exp(x), output='gradient')
                return err_, np.exp(x) * grad

            res = minimize(func, np.log(nodes), tol=tol ** 2, bounds=bounds, jac=True)

        else:
            def func(x):
                err_, grad, _ = error_fun(H, T, np.exp(x), output='gradient')
                return err_, np.exp(x) * grad

            def hess(x):
                _, grad, hessian, _ = error_fun(H, T, np.exp(x), output='hessian')
                return hessian * np.exp(x[None, :] + x[:, None]) + np.diag(grad * np.exp(x))

            res = minimize(func, np.log(nodes), tol=tol ** 2, bounds=bounds, jac=True, hess=hess, method='trust-constr')

    # post-processing, ensuring that the results are of good quality
    nodes = np.exp(res.x)
    err, weights = error_fun(H=H, T=T, nodes=nodes, output='error')
    if H > 0:
        if err > 2 * np.fmax(original_error, 1e-9):
            # return np.sqrt(np.fmax(original_error, 0)) / kernel_norm(H, T), original_nodes, original_weights
            return np.sqrt(np.fmax(original_error, 0)), original_nodes, original_weights
        # all_errors[-1] = np.sqrt(np.fmax(err, 0)) / kernel_norm(H, T)
        all_errors[-1] = np.sqrt(np.fmax(err, 0))
    else:
        if err > 0.5 * original_error:
            # return np.sqrt(np.fmax(original_error, 0)) / kernel_norm(H, T), original_nodes, original_weights
            return original_error, original_nodes, original_weights
        # all_errors[-1] = np.sqrt(np.fmax(err, 0)) / kernel_norm(H, T)
        all_errors[-1] = err
    return all_errors, nodes, weights


def optimize_error_l1(H, N, T, iterative=False, init_nodes=None, init_weights=None):
    """
    Optimizes the L^1 error with N points for the fractional kernel.
    :param H: Hurst parameter
    :param N: Number of points
    :param T: Final time, may be a numpy array (only if grad is False and fast is True)
    :param iterative: If True, starts with one node and iteratively adds nodes, while always optimizing
    :param init_nodes: May specify a starting point for the nodes
    :param init_weights: May specify a starting point for the weights
    :return: The minimal relative error together with the associated nodes and weights.
    """
    def optimize_error_given_rule(nodes_1, weights_1):
        N_ = len(nodes_1)
        coefficient = 1 / kernel_norm(H=H, T=T, p=1.)
        rule = np.log(np.concatenate((nodes_1, weights_1)))

        def func(x):
            err_, grad = error_l1(H=H, nodes=np.exp(x[:N_]), weights=np.exp(x[N_:]), T=T, method='intersections')
            return coefficient * err_

        res = minimize(func, rule, tol=1e-04)
        nodes_1, weights_1 = sort(np.exp(res.x[:N_]), np.exp(res.x[N_:]))
        return res.fun, nodes_1, weights_1

    if not iterative:
        if init_nodes is not None and init_weights is not None:
            nodes_, weights_ = init_nodes, init_weights
        else:
            nodes_, weights_ = quadrature_rule(H=H, N=N, T=T, mode='non-geometric l1')
        if len(nodes_) < N:
            nodes = np.zeros(N)
            weights = np.zeros(N)
            nodes[:len(nodes_)] = nodes_
            weights[:len(weights_)] = weights_
            for i in range(len(nodes_), N):
                nodes[i] = nodes_[-1] * 2 ** (i - len(nodes_) + 1)
                weights[i] = weights_[-1]
        else:
            nodes = nodes_[:N]
            weights = weights_[:N]

        return optimize_error_given_rule(nodes, weights)

    if init_nodes is None or init_weights is None:
        nodes, weights = quadrature_rule(H=H, N=1, T=T, mode='non-geometric l1')
        err, nodes, weights = optimize_error_given_rule(nodes, weights)
    else:
        err, nodes, weights = -1, init_nodes, init_weights

    while len(nodes) < N:
        print(len(nodes))
        nodes = np.append(nodes, 2 * nodes[-1])
        weights = np.append(weights, np.amax(weights))
        err, nodes, weights = optimize_error_given_rule(nodes, weights)

    return err, nodes, weights


def european_rule(H, N, T):
    """
    Returns a quadrature rule that is optimized for pricing European options under the rough Heston model.
    :param H: Hurst parameter
    :param N: Number of nodes
    :param T: Final time/Maturity
    :return: Nodes and weights
    """

    def optimizing_func(N_, tol_, bound_):
        if N_ == 1:
            nod = np.array([1 / T])
        else:
            nod = np.empty(N_)
            if len(last_nodes) == N_:
                nod = last_nodes
            else:
                nod[:-1] = last_nodes
                nod[-1] = bound_
        nod = nod / 1.03 ** np.fmin(np.arange(1, N_ + 1) ** 2, 100)
        return optimize_error_l2(H=H, N=N_, T=T, tol=tol_, bound=bound_, method='gradient', force_order=False,
                                 init_nodes=nod)

    if H > 0:
        _, nodes, weights = optimizing_func(N_=1, tol_=1e-06, bound_=None)
    else:
        _, nodes, weights = optimize_error_l1(H=H, N=1, T=T)
    if N == 1:
        return nodes, weights

    L_step = 1.15
    bound = np.amax(nodes) / L_step
    current_N = 1
    last_nodes = nodes

    while current_N < N:
        increase_N = 0
        L_step = 1.15

        while increase_N < 2:
            bound = bound * L_step
            error_, nodes, weights = optimizing_func(N_=current_N+1, tol_=1e-07/current_N, bound_=bound)
            p = np.argsort(nodes)
            nodes = nodes[p]
            weights = weights[p]
            if np.amin(nodes[1:] / nodes[:-1]) < 1.4 or np.abs(np.amin(weights)) < 1e-02 \
                    or np.abs(np.amin(weights[1:] / weights[:-1])) < 0.4:
                increase_N = 0
                L_step = 1.15
            elif error_ < optimizing_func(N_=current_N, tol_=1e-07/current_N, bound_=bound)[0]:
                increase_N += 1
                if L_step > 1.06:
                    L_step = 1.05
                    bound = bound / 1.15
            else:
                increase_N = 0
                L_step = 1.15

        current_N = current_N + 1
        last_nodes = nodes

    if N >= 4:
        return nodes, weights
    if N == 2:
        L_4 = bound * 2
        L_5 = bound * 3
        L_6 = bound * 4
    else:  # N == 3
        L_4 = bound
        L_5 = bound * 1.25
        L_6 = bound * 1.5
    error_4, nodes_4, weights_4 = optimizing_func(N_=N, tol_=1e-08, bound_=L_4)
    error_5, nodes_5, weights_5 = optimizing_func(N_=N, tol_=1e-08, bound_=L_5)
    error_6, nodes_6, weights_6 = optimizing_func(N_=N, tol_=1e-08, bound_=L_6)
    if error_4 <= error_5 and error_4 <= error_6:
        return nodes_4, weights_4
    if error_5 <= error_6:
        return nodes_5, weights_5
    return nodes_6, weights_6


def harms_rule(H, n, m):
    """
    The quadrature rule for fBm proposed by Harms.
    :param H: Hurst parameter
    :param n: Number of Gaussian quadrature intervals
    :param m: Degree of Gaussian quadrature
    :return: The nodes and weights of the quadrature rule
    """
    alpha_, beta_, gamma_, delta_ = H + 1/2, m - 1, 1/2 - H, H
    r = delta_ * m / (1 - alpha_ - beta_ + delta_ + m)
    xi_0 = n ** (-r / gamma_)
    xi_n = n ** (r / delta_)
    xi = xi_0 * np.exp(np.log(xi_n / xi_0) * np.linspace(0, 1, n + 1))
    return Gaussian_on_partition(H=H, m=m, partition=xi, fractional_weight=True)


def quadrature_rule(H, N, T, mode="european"):
    """
    Returns the nodes and weights of a quadrature rule for the fractional kernel with Hurst parameter H. The nodes are
    sorted in increasing order.
    :param H: Hurst parameter
    :param N: Total number of nodes
    :param T: Final time
    :param mode: The kind of quadrature rule that should be used
    :return: All the nodes and weights, in the form [node1, node2, ...], [weight1, weight2, ...]
    """
    if isinstance(T, np.ndarray):
        if N == 1:
            T = np.amin(T) ** (3 / 5) * np.amax(T) ** (2 / 5)
        elif N == 2:
            T = np.amin(T) ** (1 / 2) * np.amax(T) ** (1 / 2)
        elif N == 3:
            T = np.amin(T) ** (1 / 3) * np.amax(T) ** (2 / 3)
        elif N == 4:
            T = np.amin(T) ** (1 / 4) * np.amax(T) ** (3 / 4)
        elif N == 5:
            T = np.amin(T) ** (1 / 6) * np.amax(T) ** (5 / 6)
        elif N == 6:
            T = np.amin(T) ** (1 / 10) * np.amax(T) ** (9 / 10)
        else:
            T = np.amax(T)

    if mode == "optimized l2" or mode == "OL2":
        nodes, weights = optimize_error_l2(H=H, N=N, T=T)[1:3]
    elif mode == "optimized l1" or mode == "OL1":
        nodes, weights = optimize_error_l1(H=H, N=N, T=T, iterative=True)[1:3]
    elif mode == "european" or mode == "BL2":
        nodes, weights = european_rule(H=H, N=N, T=T)
    elif mode == "abi jaber" or mode == "AE":
        nodes, weights = AbiJaberElEuch_quadrature_rule(H=H, N=N, T=T)
    elif mode == "alfonsi" or mode == "AK":
        nodes, weights = AK_improved_rule(H=H, N=N, T=T)
    elif mode == "paper" or mode == "OLD":
        nodes, weights = Gaussian_rule(H=H, N=N, T=T, mode="old geometric observation l2")
    else:
        nodes, weights = Gaussian_rule(H=H, N=N, T=T, mode=mode)
    weights[np.logical_and(nodes < 1, np.abs(weights) > 100)] = 0
    return sort(nodes, weights)

# ---------------------------------------------------------------------------------------

# Functions taken from Christian




class kernel_frac:
    """Class representing the RL kernel."""

    def __init__(self, H, eta):
        self.H = H
        self.eta = eta
        self.eta_tilde = np.sqrt(2 * H) * eta

    def K_diag(self, Delta, N):
        """
        Return the diagonal values of calligraphic K_{i,j} as defined by Jim.
        Parameters
        ----------
        Delta : double
            Time increment for the simulation scheme, Delta = T / N.
        N : int
            Number of time steps in the simulation scheme.
        Returns
        -------
        numpy array
            The values \mathcal{K}_{j,j}(Delta), j=0, ..., N-1. Size = N.
        """
        i = np.arange(N + 1)
        # Hint: i[-N:] = (i[1],...,i[N]), i[:N] = (i[0],...,i[N-1])
        return self.eta ** 2 * Delta ** (2 * self.H) * (i[-N:] ** (2 * self.H) - i[:N] ** (2 * self.H))

    def K_0(self, Delta):
        """
        Return the value of calligraphic K_0 as defined by Jim.
        Parameters
        ----------
        Delta : double
            Time increment for the simulation scheme, Delta = T / N..
        Returns
        -------
        double
            The value \mathcal{K}_0(Delta).
        """
        return self.eta_tilde * Delta ** (self.H + 0.5) / (self.H + 0.5)


class kernel_rheston:
    """Kernel for the rHeston model, seen as a forward variance model.

    The shape of the kernel is defined in the paper of Jim and Martin, 2019."""

    def __init__(self, H, lam, zeta, eps=1e-3):
        self.alpha = H + 0.5
        self.H = H  # Is this really the "H"???
        self.lam = lam
        self.zeta = zeta
        self.eps = eps

    def _k(self, r):
        """The kernel."""
        return self.zeta * r ** (self.alpha - 1) * mittag_leffler(- self.lam * r ** self.alpha, self.alpha, self.alpha)

    def K_0(self, Delta):
        """
        Return the value of calligraphic K_0 as defined by Jim. Computed by
        computationally expensive numerical integration.
        Parameters
        ----------
        Delta : double
            Time increment for the simulation scheme, Delta = T / N..
        Returns
        -------
        double
            The value \mathcal{K}_0(Delta).
        """
        return integ.quad(lambda r: self._k(r), 0.0, Delta, epsabs=self.eps, epsrel=self.eps)[0]

    def K_diag(self, Delta, N):
        """
        Return the diagonal values of calligraphic K_{i,j} as defined by Jim.
        Computed by computationally expensive numerical integration.
        Parameters
        ----------
        Delta : double
            Time increment for the simulation scheme, Delta = T / N.
        N : int
            Number of time steps in the simulation scheme.
        Returns
        -------
        numpy array
            The values \mathcal{K}_{j,j}(Delta), j=0, ..., N-1. Size = N.
        """
        return np.array([integ.quad(lambda r: self._k(r + i * Delta) ** 2, 0.0, Delta,
                                    epsabs=self.eps, epsrel=self.eps)[0] for i in range(N)])

    def xi(self, t_grid, v0, lam, theta, eps=1e-6):
        """
        Compute the forward variance curve for the rough Heston model.

        Note that the forward variance curve is a constant if v0 is equal
        to theta.
        Parameters
        ----------
        t_grid : numpy array or scalar.
            Time grid or value, along which the forward variance curve is
            requested. Assumed to be non-negative.
        v0 : scalar.
            Initial variance.
        lam : scalar.
            Speed of mean reversion.
        theta : scalar
            Long-term mean variance.
        eps : positive scalar, optional
            Accuracy target for the quadrature method. The default is 1e-6.
        Returns
        -------
        Numpy array.
            The forward variance curve.
        """
        if np.isclose(v0, theta, rtol=eps):
            return np.full_like(t_grid, v0)

        t = np.unique(np.append(0.0, t_grid))

        # integrate between t_i and t_{i+1}
        def f(i):
            return integ.quad(self._k, t[i], t[i + 1], epsabs=eps, epsrel=eps)[0]

        int_k = np.array([f(i) for i in range(len(t) - 1)])
        if isinstance(t_grid, np.ndarray) and t_grid[0] == 0.0:
            int_k = np.append(0.0, int_k)
        elif t_grid == 0.0:
            int_k = np.append(0.0, int_k)

        # Note that the forward variance curve uses the kernel without zeta.
        return v0 + self.lam * (theta - v0) * np.cumsum(int_k) / self.zeta



"""
implementation of gaussian mixture pricer and calibration
"""
import numpy as np
import matplotlib.pyplot as plt
from dataclasses import dataclass
from scipy.optimize import minimize
from numba.typed import List
from typing import Tuple
from enum import Enum

# sv 
import stochvolmodels.pricers.analytic.tdist as td
from stochvolmodels.utils.funcs import to_flat_np_array, timer
from stochvolmodels.pricers.model_pricer import ModelParams, ModelPricer
from stochvolmodels.utils.config import VariableType

# data
from stochvolmodels.data.option_chain import OptionChain


@dataclass
class TdistParams(ModelParams):
    drift: float
    vol: float
    nu: float
    ttm: float  # ttm is important as all params are fixed to this ttm, it is not part of calibration


class TdistPricer(ModelPricer):

    def price_chain(self, option_chain: OptionChain, params: TdistParams, **kwargs) -> np.ndarray:
        """
        implementation of generic method price_chain using heston wrapper for tdist prices
        """
        model_prices_ttms = tdist_vanilla_chain_pricer(drift=params.drift,
                                                       vol=params.vol,
                                                       nu=params.nu,
                                                       ttms=option_chain.ttms,
                                                       forwards=option_chain.forwards,
                                                       strikes_ttms=option_chain.strikes_ttms,
                                                       optiontypes_ttms=option_chain.optiontypes_ttms,
                                                       discfactors=option_chain.discfactors)

        return model_prices_ttms

    def model_mc_price_chain(self, option_chain: OptionChain, params: TdistParams,
                             nb_path: int = 100000,
                             variable_type: VariableType = VariableType.LOG_RETURN,
                             **kwargs
                             ) -> (List[np.ndarray], List[np.ndarray]):
        raise NotImplementedError

    @timer
    def calibrate_model_params_to_chain_slice(self,
                                              option_chain: OptionChain,
                                              params0: TdistParams = None,
                                              is_vega_weighted: bool = True,
                                              is_unit_ttm_vega: bool = False,
                                              **kwargs
                                              ) -> TdistParams:
        """
        implementation of model calibration interface
        fit: TdistParams
        nb: always use option_chain with one slice because we need martingale condition per slice
        """
        ttms = option_chain.ttms
        if len(ttms) > 1:
            raise NotImplementedError(f"cannot calibrate to multiple slices")
        ttm = ttms[0]
        rf_rate = option_chain.discount_rates[0]

        # p0 = (gmm_weights, gmm_mus, gmm_vols)
        if params0 is not None:
            p0 = np.array([params0.vol, params0.nu])
        else:
            p0 = np.array([0.2, 3.0])

        vol_bounds = [(0.05, 10.0)]
        nu_bounds = [(2.01, 20.0)]
        bounds = np.concatenate((vol_bounds, nu_bounds))

        x, y = option_chain.get_chain_data_as_xy()
        market_vols = to_flat_np_array(y)  # market mid quotes
        if is_vega_weighted:
            vegas_ttms = option_chain.get_chain_vegas(is_unit_ttm_vega=is_unit_ttm_vega)
            vegas_ttms = [vegas_ttm/sum(vegas_ttm) for vegas_ttm in vegas_ttms]
            weights = to_flat_np_array(vegas_ttms)
        else:
            weights = np.ones_like(market_vols)

        def parse_model_params(pars: np.ndarray) -> TdistParams:
            vol = pars[0]
            nu = pars[1]
            drift = td.imply_drift_tdist(rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
            return TdistParams(vol=vol, nu=nu, drift=drift, ttm=ttm)

        def objective(pars: np.ndarray, args: np.ndarray) -> float:
            params = parse_model_params(pars=pars)
            model_vols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params)
            resid = np.nansum(weights * np.square(to_flat_np_array(model_vols) - market_vols))
            return resid

        options = {'disp': True, 'ftol': 1e-10, 'maxiter': 500}
        res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)
        fit_params = parse_model_params(pars=res.x)

        return fit_params

    @timer
    def calibrate_model_params_to_chain(self,
                                        option_chain: OptionChain,
                                        is_vega_weighted: bool = True,
                                        is_unit_ttm_vega: bool = False,
                                        **kwargs
                                        ) -> List[str, TdistParams]:
        """
        model params are fitted per slice
        need to splic chain to slices
        """
        fit_params = {}
        params0 = None
        for ids_ in option_chain.ids:
            option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[ids_])
            params0 = self.calibrate_model_params_to_chain_slice(option_chain=option_chain0,
                                                                 params0=params0,
                                                                 is_vega_weighted=is_vega_weighted,
                                                                 is_unit_ttm_vega=is_unit_ttm_vega,
                                                                 **kwargs)
            fit_params[ids_] = params0
        return fit_params


def tdist_vanilla_chain_pricer(vol: float,
                               nu: float,
                               drift: float,
                               ttms: np.ndarray,
                               forwards: np.ndarray,
                               strikes_ttms: Tuple[np.ndarray, ...],
                               optiontypes_ttms: Tuple[np.ndarray, ...],
                               discfactors: np.ndarray,
                               ) -> np.ndarray:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    # outputs as numpy lists
    model_prices_ttms = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, discfactors, strikes_ttms,
                                                                      optiontypes_ttms):
        option_prices_ttm = td.compute_vanilla_price_tdist(spot=forward*discfactor,
                                                           strikes=strikes_ttm,
                                                           ttm=ttm,
                                                           vol=vol,
                                                           nu=nu,
                                                           optiontypes=optiontypes_ttm,
                                                           rf_rate=drift,
                                                           is_compute_risk_neutral_mu=False  # drift is already adjusted
                                                           )
        model_prices_ttms.append(option_prices_ttm)

    return model_prices_ttms


class UnitTests(Enum):
    CALIBRATOR = 1


def run_unit_test(unit_test: UnitTests):

    import seaborn as sns
    from stochvolmodels.utils import plots as plot
    import stochvolmodels.data.test_option_chain as chains

    if unit_test == UnitTests.CALIBRATOR:
        # option_chain = chains.get_btc_test_chain_data()
        option_chain = chains.get_spy_test_chain_data()
        # option_chain = chains.get_gld_test_chain_data()

        tdist_pricer = TdistPricer()
        fit_params = tdist_pricer.calibrate_model_params_to_chain(option_chain=option_chain)

        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(2, 2, figsize=(14, 12), tight_layout=True)
            axs = plot.to_flat_list(axs)

        for idx, (key, params) in enumerate(fit_params.items()):
            print(f"{key}: {params}")
            option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[key])
            tdist_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain0, params=params, axs=[axs[idx]])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CALIBRATOR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
run few unit test to illustrate implementation of log-normal sv model analytics
"""

import numpy as np
import matplotlib.pyplot as plt
from enum import Enum
import stochvolmodels as sv
from stochvolmodels import LogSVPricer, LogSvParams, OptionChain, LogsvModelCalibrationType


class UnitTests(Enum):
    COMPUTE_MODEL_PRICES = 1
    PLOT_MODEL_IMPLIED_VOLS = 2
    PLOT_MODEL_VOLS_IN_PARAMS = 3
    COMPARE_MODEL_VOLS_TO_MC = 4
    PLOT_FIT_TO_BITCOIN_OPTION_CHAIN = 5
    CALIBRATE_MODEL_TO_BTC_OPTIONS = 6
    MC_WITH_FIXED_RANDOMS = 7
    CALIBRATE_MODEL_TO_BTC_OPTIONS_WITH_MC = 8
    ROUGH_MC_WITH_FIXED_RANDOMS = 9
    BENCHM_ROUGH_PRICER = 10


def run_unit_test(unit_test: UnitTests):

    # instance of pricer
    logsv_pricer = LogSVPricer()
    
    # define model params    
    params = LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=5.0, beta=0.2, volvol=2.0)

    if unit_test == UnitTests.COMPUTE_MODEL_PRICES:
        # 1. one price
        model_price, vol = logsv_pricer.price_vanilla(params=params,
                                                      ttm=0.25,
                                                      forward=1.0,
                                                      strike=1.0,
                                                      optiontype='C')
        print(f"price={model_price:0.4f}, implied vol={vol: 0.2%}")

        # 2. prices for slices
        model_prices, vols = logsv_pricer.price_slice(params=params,
                                                      ttm=0.25,
                                                      forward=1.0,
                                                      strikes=np.array([0.9, 1.0, 1.1]),
                                                      optiontypes=np.array(['P', 'C', 'C']))
        print([f"{p:0.4f}, implied vol={v: 0.2%}" for p, v in zip(model_prices, vols)])

        # 3. prices for option chain with uniform strikes
        option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.9, 1.1, 3))
        model_prices, vols = logsv_pricer.compute_chain_prices_with_vols(option_chain=option_chain, params=params)
        print(model_prices)
        print(vols)

    elif unit_test == UnitTests.PLOT_MODEL_IMPLIED_VOLS:
        # define uniform option chain
        option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.5, 1.5, 21))
        logsv_pricer.plot_model_ivols(option_chain=option_chain,
                                      params=params)

    elif unit_test == UnitTests.PLOT_MODEL_VOLS_IN_PARAMS:

        # define uniform option chain
        option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.5, 1.5, 21))

        # define parameters for bootstrap
        params_dict = {'kappa2=5': LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=5.0, beta=0.2, volvol=2.0),
                       'kappa2=10': LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=10.0, beta=0.2, volvol=2.0)}

        # get slice for illustration
        option_slice = option_chain.get_slice(id='1m')
        logsv_pricer.plot_model_slices_in_params(option_slice=option_slice,
                                                 params_dict=params_dict)

    elif unit_test == UnitTests.COMPARE_MODEL_VOLS_TO_MC:
        btc_option_chain = sv.get_btc_test_chain_data()
        uniform_chain_data = OptionChain.to_uniform_strikes(obj=btc_option_chain, num_strikes=31)
        btc_calibrated_params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8609, kappa2=4.7940, beta=0.1988, volvol=2.3694)

        logsv_pricer.plot_model_ivols_vs_mc(option_chain=uniform_chain_data,
                                            params=btc_calibrated_params,
                                            nb_path=100000)

        logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=uniform_chain_data,
                                                           params=btc_calibrated_params,
                                                           nb_path=100000)

    elif unit_test == UnitTests.PLOT_FIT_TO_BITCOIN_OPTION_CHAIN:
        btc_option_chain = sv.get_btc_test_chain_data()
        btc_calibrated_params = LogSvParams(sigma0=0.8327, theta=1.0139, kappa1=4.8609, kappa2=4.7940, beta=0.1988, volvol=2.3694)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                                 params=btc_calibrated_params)

    elif unit_test == UnitTests.MC_WITH_FIXED_RANDOMS:
        btc_option_chain = sv.get_btc_test_chain_data()
        W0s, W1s, dts = sv.get_randoms_for_chain_valuation(ttms=btc_option_chain.ttms,
                                                           nb_path=10000,
                                                           nb_steps_per_year=360,
                                                           seed=10)
        print(dts)
        params0 = LogSvParams(sigma0=0.8, theta=1.0, kappa1=2.21, kappa2=2.18, beta=0.15, volvol=2.0)
        vol_backbone_etas = params.get_vol_backbone_etas(ttms=btc_option_chain.ttms)
        args = dict(ttms=btc_option_chain.ttms,
                    forwards=btc_option_chain.forwards,
                    discfactors=btc_option_chain.discfactors,
                    strikes_ttms=btc_option_chain.strikes_ttms,
                    optiontypes_ttms=btc_option_chain.optiontypes_ttms,
                    W0s=W0s,
                    W1s=W1s,
                    dts=dts,
                    v0=params0.sigma0,
                    theta=params0.theta,
                    kappa1=params0.kappa1,
                    kappa2=params0.kappa2,
                    beta=params0.beta,
                    volvol=params0.volvol,
                    vol_backbone_etas=vol_backbone_etas)
        option_prices_ttm, option_std_ttm = sv.logsv_mc_chain_pricer_fixed_randoms(**args)
        print(option_prices_ttm)


    elif unit_test == UnitTests.ROUGH_MC_WITH_FIXED_RANDOMS:
        btc_option_chain = sv.get_btc_test_chain_data()
        Z0, Z1, grid_ttms = sv.get_randoms_for_rough_vol_chain_valuation(ttms=btc_option_chain.ttms,
                                                                             nb_path=10000,
                                                                             nb_steps_per_year=360,
                                                                             seed=10)
        params0 = LogSvParams(sigma0=0.8, theta=1.0, kappa1=2.21, kappa2=0.0, beta=0.15, volvol=2.0)
        params0.H = 0.3
        params0.approximate_kernel(T=btc_option_chain.ttms[-1], N=3)

        option_prices_ttm, option_std_ttm = sv.rough_logsv_mc_chain_pricer_fixed_randoms(ttms=btc_option_chain.ttms,
                                                                                         forwards=btc_option_chain.forwards,
                                                                                         discfactors=btc_option_chain.discfactors,
                                                                                         strikes_ttms=btc_option_chain.strikes_ttms,
                                                                                         optiontypes_ttms=btc_option_chain.optiontypes_ttms,
                                                                                         Z0=Z0,
                                                                                         Z1=Z1,
                                                                                         sigma0=params0.sigma0,
                                                                                         theta=params0.theta,
                                                                                         kappa1=params0.kappa1,
                                                                                         kappa2=params0.kappa2,
                                                                                         beta=params0.beta,
                                                                                         orthog_vol=params0.volvol,
                                                                                         weights=params0.weights,
                                                                                         nodes=params0.nodes,
                                                                                         timegrids=grid_ttms)
        print(option_prices_ttm)

    elif unit_test == UnitTests.BENCHM_ROUGH_PRICER:
        btc_option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                                     ids=np.array(['1m', '3m']),
                                                     strikes=np.linspace(0.5, 1.5, 21))
        params0 = LogSvParams(sigma0=0.8, theta=1.0, kappa1=2.21, kappa2=0.0, beta=0.15, volvol=2.0)
        nb_path = 100000
        H = 0.3
        N = 3

        def rough_vol():
            params1 = LogSvParams.copy(params0)
            params1.H = 0.3
            params1.approximate_kernel(T=btc_option_chain.ttms[-1], N=N)

            Z0, Z1, grid_ttms = sv.get_randoms_for_rough_vol_chain_valuation(ttms=btc_option_chain.ttms,
                                                                             nb_path=nb_path,
                                                                             nb_steps_per_year=360,
                                                                             seed=10)


            option_prices_ttm, option_std_ttm = sv.rough_logsv_mc_chain_pricer_fixed_randoms(ttms=btc_option_chain.ttms,
                                                                                             forwards=btc_option_chain.forwards,
                                                                                             discfactors=btc_option_chain.discfactors,
                                                                                             strikes_ttms=btc_option_chain.strikes_ttms,
                                                                                             optiontypes_ttms=btc_option_chain.optiontypes_ttms,
                                                                                             Z0=Z0,
                                                                                             Z1=Z1,
                                                                                             sigma0=params0.sigma0,
                                                                                             theta=params0.theta,
                                                                                             kappa1=params0.kappa1,
                                                                                             kappa2=params0.kappa2,
                                                                                             beta=params0.beta,
                                                                                             orthog_vol=params0.volvol,
                                                                                             weights=params1.weights,
                                                                                             nodes=params1.nodes,
                                                                                             timegrids=grid_ttms)
            model_ivols_ttms = btc_option_chain.compute_model_ivols_from_chain_data(option_prices_ttm)
            return model_ivols_ttms

        def regular_vol():
            W0s, W1s, dts = sv.get_randoms_for_chain_valuation(ttms=btc_option_chain.ttms,
                                                               nb_path=nb_path,
                                                               nb_steps_per_year=360,
                                                               seed=10)
            vol_backbone_etas = params.get_vol_backbone_etas(ttms=btc_option_chain.ttms)
            args = dict(ttms=btc_option_chain.ttms,
                        forwards=btc_option_chain.forwards,
                        discfactors=btc_option_chain.discfactors,
                        strikes_ttms=btc_option_chain.strikes_ttms,
                        optiontypes_ttms=btc_option_chain.optiontypes_ttms,
                        W0s=W0s,
                        W1s=W1s,
                        dts=dts,
                        v0=params0.sigma0,
                        theta=params0.theta,
                        kappa1=params0.kappa1,
                        kappa2=params0.kappa2,
                        beta=params0.beta,
                        volvol=params0.volvol,
                        vol_backbone_etas=vol_backbone_etas)
            option_prices_ttm, option_std_ttm = sv.logsv_mc_chain_pricer_fixed_randoms(**args)
            model_ivols_ttms = btc_option_chain.compute_model_ivols_from_chain_data(option_prices_ttm)

            return model_ivols_ttms

        ivols_rough_logsv = rough_vol()
        ivols_logsv = regular_vol()

        nb_slices = btc_option_chain.ttms.size
        fig, axs = plt.subplots(1, nb_slices, figsize=(4*nb_slices, 3), tight_layout=True)

        for i in range(nb_slices):
            ax = axs[i] if nb_slices>1 else axs
            ax.plot(btc_option_chain.strikes_ttms[i], ivols_logsv[i], label="LOG_SV", marker="*")
            ax.plot(btc_option_chain.strikes_ttms[i], ivols_rough_logsv[i], label="ROUGH_LOG_SV", marker="o")
            ax.set_title("Expiry: " + btc_option_chain.ids[i])
            ax.legend()
        fig.suptitle(f"Conventional LogSV model vs Rough LogSV, H={H:.2f} via {N}f Markovian approximation",
                     color="darkblue")

    elif unit_test == UnitTests.CALIBRATE_MODEL_TO_BTC_OPTIONS:
        btc_option_chain = sv.get_btc_test_chain_data()
        params0 = LogSvParams(sigma0=0.8, theta=1.0, kappa1=2.21, kappa2=2.18, beta=0.15, volvol=2.0)
        btc_calibrated_params = logsv_pricer.calibrate_model_params_to_chain(option_chain=btc_option_chain,
                                                                             params0=params0,
                                                                             model_calibration_type=LogsvModelCalibrationType.PARAMS4,
                                                                             constraints_type=sv.ConstraintsType.INVERSE_MARTINGALE)
        print(btc_calibrated_params)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                                 params=btc_calibrated_params)

    elif unit_test == UnitTests.CALIBRATE_MODEL_TO_BTC_OPTIONS_WITH_MC:
        btc_option_chain = sv.get_btc_test_chain_data()
        params0 = LogSvParams(sigma0=0.8, theta=1.0, kappa1=2.21, kappa2=0.0, beta=0.15, volvol=2.0)
        params0.H = 0.3
        params0.approximate_kernel(T=btc_option_chain.ttms[-1], N=3)
        btc_calibrated_params = logsv_pricer.calibrate_model_params_to_chain(option_chain=btc_option_chain,
                                                                             params0=params0,
                                                                             model_calibration_type=LogsvModelCalibrationType.PARAMS4,
                                                                             constraints_type=sv.ConstraintsType.INVERSE_MARTINGALE,
                                                                             calibration_engine=sv.CalibrationEngine.ROUGH_MC,
                                                                             nb_path=100000,
                                                                             seed=7)
        print(btc_calibrated_params)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                                 params=btc_calibrated_params)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.BENCHM_ROUGH_PRICER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
from typing import Tuple, Union


def de_pricer(ff, ff_transf) -> Tuple[np.ndarray, np.ndarray]:
    eps0 = 1e-6
    h = 0.5
    eps = 1e-6
    Nmax = 12.0
    maxlev = 7

    s = func(ff, 0.0)
    # add terms for k > 0
    n1, s = trunc_index(ff, h2=h, delta=1, s=s, Nmax=Nmax, eps0=eps0)
    # add terms for k < 0
    n2, s = trunc_index(ff, h2=-h, delta=1, s=s, Nmax=Nmax, eps0=eps0)
    # level 0 estimate
    model_prices_ttm_prev = h * s
    model_ivs_ttm_prev = ff_transf(model_prices_ttm_prev)[1]
    m = 0
    err_ivol = 1
    model_prices_ttm = None
    model_ivs_ttm = None
    for m in np.arange(1.0, maxlev):
        h = h / 2.0
        # add terms in refined mesh for k > 0
        s1 = part_sum(ff, h2=h, delta=2, N=n1)
        # add terms in refined mesh for k < 0
        s2 = part_sum(ff, h2=-h, delta=2, N=n2)
        model_prices_ttm = 0.5 * model_prices_ttm_prev + h * (s1 + s2)
        # print(f"model_prices: {model_prices_ttm[0]}")
        model_ivs_ttm = ff_transf(model_prices_ttm)[1]
        err_ivol = np.linalg.norm(model_ivs_ttm - model_ivs_ttm_prev)
        rel_diff = np.linalg.norm(model_prices_ttm - model_prices_ttm_prev) <= eps * np.linalg.norm(model_prices_ttm)

        if rel_diff or err_ivol <= 1e-6:
            break
        else:
            # update
            model_prices_ttm_prev = model_prices_ttm
            model_ivs_ttm_prev = model_ivs_ttm
            n1 = 2 * n1
            n2 = 2 * n2
    # we calculate prices of capped payoff => need to transform into call option prices
    model_prices_ttm = ff_transf(model_prices_ttm)[0]
    if m == maxlev - 1 and err_ivol > 1e-6:
        # raise ValueError(f"errvol = {err_ivol:.6f}, error is above tolerance")
        # print(f"errvol = {err_ivol:.6f}, error is above tolerance")
        pass
    return model_prices_ttm, model_ivs_ttm


def func(ff, x: Union[float, np.ndarray]) -> np.ndarray:
    """calculate the term w_k * f(x_k) of the DE scheme
    x must equal kh for some k"""
    if isinstance(x, float):
        x = np.array([x])
    half_pi = 0.5 * np.pi
    exp_x = np.exp(x)
    sinh_x = 0.5 * (exp_x - 1.0 / exp_x)
    cosh_x = 0.5 * (exp_x + 1.0 / exp_x)
    exp_sinh_x = np.exp(half_pi * sinh_x)
    w_k = half_pi * cosh_x * exp_sinh_x
    x_k = exp_sinh_x
    val = (ff(x_k).T * w_k).T
    return val


def part_sum(ff, h2: float, delta: int, N: int) -> float:
    s = 0.0
    func_vals = func(ff, h2 + np.arange(0.0, N, 1.0) * delta * h2)
    for idx, func_val in enumerate(func_vals):
        s = s + func_vals[idx]
    return s


def trunc_index(ff,
                h2: float,
                delta: int,
                s: np.ndarray,
                Nmax: float,
                eps0: float) -> (int, np.ndarray):
    x = h2
    k = 1
    for k in np.arange(1.0, Nmax):
        xi = func(ff, x)
        s = s + xi
        if np.all(np.linalg.norm(xi, axis=0) <= eps0 * np.linalg.norm(s, axis=0)):
            break
        x = x + delta * h2
    return k, s





"""
implementation of log sv params
"""
import numpy as np
import pandas as pd
from numpy import linalg as la
from dataclasses import dataclass, asdict
from typing import Optional, Dict, Any

from stochvolmodels import VariableType, find_nearest
from stochvolmodels.pricers.model_pricer import ModelParams
from stochvolmodels.pricers.rough_logsv.RoughKernel import european_rule


@dataclass
class LogSvParams(ModelParams):
    """
    Implementation of model params class
    """
    sigma0: float = 0.2
    theta: float = 0.2
    kappa1: float = 1.0
    kappa2: Optional[float] = 2.5  # Optional is mapped to self.kappa1 / self.theta
    beta: float = -1.0
    volvol: float = 1.0
    vol_backbone: pd.Series = None
    H: float = 0.5
    weights: np.ndarray = None
    nodes: np.ndarray = None

    def __post_init__(self):
        if self.kappa2 is None:
            self.kappa2 = self.kappa1 / self.theta
        assert 1e-4 < self.H <= 0.5

    def approximate_kernel(self, T: float, N: int):
        assert 1 <= N <= 5  # not keen to use large N
        if N > 1 and self.H<0.49:
            self.nodes, self.weights = european_rule(self.H, N, T)
        else:
            self.weights = np.array([1.0])
            self.nodes = np.array([1e-3])


    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def to_str(self) -> str:
        return f"sigma0={self.sigma0:0.2f}, theta={self.theta:0.2f}, kappa1={self.kappa1:0.2f}, kappa2={self.kappa2:0.2f}, " \
               f"beta={self.beta:0.2f}, volvol={self.volvol:0.2f}"

    def set_vol_backbone(self, vol_backbone: pd.Series) -> None:
        self.vol_backbone = vol_backbone

    def get_vol_backbone_eta(self, tau: float) -> float:
        if self.vol_backbone is not None:
            nearest_tau = find_nearest(a=self.vol_backbone.index.to_numpy(), value=tau, is_equal_or_largest=True)
            vol_backbone_eta = self.vol_backbone.loc[nearest_tau]
        else:
            vol_backbone_eta = 1.0
        return vol_backbone_eta

    def get_vol_backbone_etas(self, ttms: np.ndarray) -> np.ndarray:
        if self.vol_backbone is not None:
            vol_backbone_etas = np.ones_like(ttms)
            for idx, tau in enumerate(ttms):
                nearest_tau = find_nearest(a=self.vol_backbone.index.to_numpy(), value=tau, is_equal_or_largest=True)
                vol_backbone_etas[idx] = self.vol_backbone.loc[nearest_tau]
        else:
            vol_backbone_etas = np.ones_like(ttms)
        return vol_backbone_etas

    @property
    def kappa(self) -> float:
        return self.kappa1+self.kappa2*self.theta

    @property
    def theta2(self) -> float:
        return self.theta*self.theta

    @property
    def vartheta2(self) -> float:
        return self.beta*self.beta + self.volvol*self.volvol

    @property
    def gamma(self) -> float:
        """
        assume kappa2 = kappa1 / theta
        """
        return self.kappa1 / self.theta

    @property
    def eta(self) -> float:
        """
        assume kappa2 = kappa1 / theta
        """
        return self.kappa1 * self.theta / self.vartheta2 - 1.0

    def get_x_grid(self, ttm: float = 1.0, n_stdevs: float = 3.0, n: int = 200) -> np.ndarray:
        """
        spacial grid to compute density of x
        """
        sigma_t = np.sqrt(ttm * 0.5 * (np.square(self.sigma0) + np.square(self.theta)))
        drift = - 0.5*sigma_t*sigma_t
        stdev = (n_stdevs+1)*sigma_t
        return np.linspace(-stdev+drift, stdev+drift, n)

    def get_sigma_grid(self, ttm: float = 1.0, n_stdevs: float = 3.0, n: int = 200) -> np.ndarray:
        """
        spacial grid to compute density of sigma
        """
        sigma_t = np.sqrt(0.5*(np.square(self.sigma0) + np.square(self.theta)))
        vvol = 0.5*np.sqrt(self.vartheta2*ttm)
        return np.linspace(0.0, sigma_t+n_stdevs*vvol, n)

    def get_qvar_grid(self, ttm: float = 1.0, n_stdevs: float = 3.0, n: int = 200) -> np.ndarray:
        """
        spacial grid to compute density of i
        """
        sigma_t = np.sqrt(ttm * (np.square(self.sigma0) + np.square(self.theta)))
        vvol = np.sqrt(self.vartheta2)*ttm
        return np.linspace(0.0, sigma_t+n_stdevs*vvol, n)

    def get_variable_space_grid(self, variable_type: VariableType = VariableType.LOG_RETURN,
                                ttm: float = 1.0,
                                n_stdevs: float = 3,
                                n: int = 200
                                ) -> np.ndarray:
        if variable_type == VariableType.LOG_RETURN:
            return self.get_x_grid(ttm=ttm, n_stdevs=n_stdevs, n=n)
        if variable_type == VariableType.SIGMA:
            return self.get_sigma_grid(ttm=ttm, n_stdevs=n_stdevs, n=n)
        elif variable_type == VariableType.Q_VAR:
            return self.get_qvar_grid(ttm=ttm, n_stdevs=n_stdevs, n=n)
        else:
            raise NotImplementedError

    def get_vol_moments_lambda(self,
                               n_terms: int = 4
                               ) -> np.ndarray:

        kappa2 = self.kappa2
        kappa = self.kappa
        vartheta2 = self.vartheta2
        theta = self.theta
        theta2 = self.theta2

        def c(n: int) -> float:
            return 0.5 * vartheta2 * n * (n - 1.0)

        lambda_m = np.zeros((n_terms, n_terms))
        lambda_m[0, 0] = -kappa
        lambda_m[0, 1] = -kappa2
        lambda_m[1, 0] = 2.0*c(2) * theta
        lambda_m[1, 1] = c(2) - 2.0*kappa
        lambda_m[1, 2] = -2.0*kappa2

        for n_ in np.arange(2, n_terms):
            n = n_ + 1  # n_ is array counter, n is formula counter
            c_n = c(n)
            lambda_m[n_, n_ - 2] = c_n * theta2
            lambda_m[n_, n_ - 1] = 2.0 * c_n * theta
            lambda_m[n_, n_] = c_n - n*kappa
            if n_ + 1 < n_terms:
                lambda_m[n_, n_ + 1] = -n*kappa2

        return lambda_m

    def assert_vol_moments_stability(self, n_terms: int = 4):
        lambda_m = self.get_vol_moments_lambda(n_terms=n_terms)
        w, v = la.eig(lambda_m)
        cond = np.all(np.real(w)<0.0)
        print(f"vol moments stable = {cond}")

    def print_vol_moments_stability(self, n_terms: int = 4) -> None:
        def c(n: int) -> float:
            return 0.5 * self.vartheta2 * n * (n - 1.0)

        cond_m2 = c(2) - 2.0*self.kappa
        print(f"con2:\n{cond_m2}")
        cond_m3 = c(3) - 3.0*self.kappa
        print(f"con3:\n{cond_m3}")
        cond_m4 = c(4) - 4.0*self.kappa
        print(f"cond4:\n{cond_m4}")

        lambda_m = self.get_vol_moments_lambda(n_terms=n_terms)
        print(f"lambda_m:\n{lambda_m}")

        w, v = la.eig(lambda_m)
        print(f"eigenvalues w:\n{w}")
        print(f"vol moments stable = {np.all(np.real(w)<0.0)}")



"""
test option valuation using moment generating function analytics for Black-Scholes-Merton model
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple
from enum import Enum

import stochvolmodels.utils.mgf_pricer as mgfp
from stochvolmodels.pricers.analytic.bsm import infer_bsm_ivols_from_model_chain_prices
from stochvolmodels.utils.config import VariableType


def compute_normal_mgf_grid(ttm: float,
                            vol: float,
                            is_spot_measure: bool = True
                            ) -> Tuple[np.ndarray, np.ndarray]:
    """
    for normal: mgf = exp( 0.5*PHI * (PHI + 1.0)*sigma^2*T )
    add two columns to make compatible with @ for numba in slice_pricer_with_a_grid
    """
    phi_grid = mgfp.get_phi_grid(is_spot_measure=is_spot_measure)
    if is_spot_measure:  # use 1 for spot measure
        alpha = 1.0
    else:
        alpha = -1.0
    log_mgf_grid = 0.5 * phi_grid * (phi_grid + alpha) * (ttm * vol * vol)
    return log_mgf_grid, phi_grid


def compute_normal_mgf_psi_grid(ttm: float,
                                vol: float,
                                is_spot_measure: bool = True
                                ) -> Tuple[np.ndarray, np.ndarray]:
    psi_grid = mgfp.get_psi_grid(is_spot_measure=is_spot_measure)
    log_mgf_grid = -psi_grid * (ttm * vol * vol)
    return log_mgf_grid, psi_grid


def bsm_slice_pricer(ttm: float,
                     forward: float,
                     vol: float,
                     strikes: np.ndarray,
                     optiontypes: np.ndarray,
                     variable_type: VariableType = VariableType.LOG_RETURN,
                     is_spot_measure: bool = True
                     ) -> Tuple[np.ndarray, np.ndarray]:

    if variable_type == VariableType.LOG_RETURN:
        log_mgf_grid, phi_grid = compute_normal_mgf_grid(ttm=ttm, vol=vol, is_spot_measure=is_spot_measure)
        bsm_prices = mgfp.vanilla_slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                             phi_grid=phi_grid,
                                                             forward=forward,
                                                             strikes=strikes,
                                                             optiontypes=optiontypes,
                                                             is_spot_measure=is_spot_measure)
        bsm_ivols = infer_bsm_ivols_from_model_chain_prices(ttms=np.array([ttm]),
                                                            forwards=np.array([forward]),
                                                            discfactors=np.array([1.0]),
                                                            strikes_ttms=(strikes,),
                                                            optiontypes_ttms=(optiontypes,),
                                                            model_prices_ttms=(bsm_prices,))
    elif variable_type == VariableType.Q_VAR:
        log_mgf_grid, psi_grid = compute_normal_mgf_psi_grid(ttm=ttm, vol=vol, is_spot_measure=is_spot_measure)
        bsm_prices = mgfp.slice_qvar_pricer_with_a_grid(log_mgf_grid=log_mgf_grid,
                                                        psi_grid=psi_grid,
                                                        ttm=ttm,
                                                        strikes=strikes,
                                                        optiontypes=optiontypes,
                                                        is_spot_measure=is_spot_measure,
                                                        forward=forward)
        bsm_ivols = np.zeros_like(bsm_prices)
    else:
        raise ValueError(f"not implemented")

    return bsm_prices, bsm_ivols


def compare_spot_and_inverse_options():
    ttm = 1.0
    forward = 1.0
    vol = 1.0
    strikes = np.linspace(0.5, 5.0, 19)
    optiontypes = np.full(strikes.shape, 'C')
    optiontypes_inverse = np.full(strikes.shape, 'IC')

    # spot
    bsm_prices_spot, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                  optiontypes=optiontypes,
                                                  is_spot_measure=True)
    bsm_prices_inverse, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                     optiontypes=optiontypes_inverse,
                                                     is_spot_measure=False)

    bsm_prices_spot = pd.Series(bsm_prices_spot, index=strikes, name='spot')
    bsm_prices_inverse = pd.Series(bsm_prices_inverse, index=strikes, name='inverse')

    prices = pd.concat([bsm_prices_spot, bsm_prices_inverse], axis=1)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 4.0), tight_layout=True)
        sns.lineplot(data=prices, ax=ax)


def compare_spot_and_inverse_qvar_options():
    ttm = 1.0
    forward = 1.0
    vol = 1.0
    strikes = np.linspace(0.5, 5.0, 19)
    optiontypes = np.full(strikes.shape, 'C')
    optiontypes_inverse = np.full(strikes.shape, 'IC')

    # spot
    bsm_prices_spot, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                  optiontypes=optiontypes,
                                                  variable_type=VariableType.Q_VAR,
                                                  is_spot_measure=True)
    bsm_prices_inverse, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes,
                                                     optiontypes=optiontypes_inverse,
                                                     variable_type=VariableType.Q_VAR,
                                                     is_spot_measure=False)

    bsm_prices_spot = pd.Series(bsm_prices_spot, index=strikes, name='spot')
    bsm_prices_inverse = pd.Series(bsm_prices_inverse, index=strikes, name='inverse')

    prices = pd.concat([bsm_prices_spot, bsm_prices_inverse], axis=1)
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 4.0), tight_layout=True)
        sns.lineplot(data=prices, ax=ax)


class UnitTests(Enum):
    BSM_SLICE_PRICER = 1
    SPOT_INVERSE_COMP = 2
    SPOT_INVERSE_QVAR_COMP = 3


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.BSM_SLICE_PRICER:
        ttm = 1.0
        forward = 1.0
        vol = 1.0
        strikes = np.linspace(0.5, 5.0, 19)
        optiontypes = np.full(strikes.shape, 'C')
        bsm_prices, bsm_ivols = bsm_slice_pricer(ttm=ttm, forward=forward, vol=vol, strikes=strikes, optiontypes=optiontypes)
        print(bsm_prices)
        print(bsm_ivols)

    elif unit_test == UnitTests.SPOT_INVERSE_COMP:
        compare_spot_and_inverse_options()

    elif unit_test == UnitTests.SPOT_INVERSE_QVAR_COMP:
        compare_spot_and_inverse_qvar_options()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.BSM_SLICE_PRICER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
